"""Provides general utility methods that get used during the course of analysis."""

import itertools
from operator import itemgetter

import xarray as xr

from .funcutils import *
from .normalize_datatype import *
from .xarray import *
from .region import *
from .dict import *
from .attrs import *
from .collections import *
from .string import *


def enumerate_dataarray(arr: xr.DataArray):
    """Iterates through each coordinate location on n dataarray. Should merge to xarray_extensions."""
    for coordinate in itertools.product(*[arr.coords[d] for d in arr.dims]):
        zip_location = dict(zip(arr.dims, (float(f) for f in coordinate)))
        yield zip_location, arr.loc[zip_location].values.item()


def arrange_by_indices(items, indices):
    """Arranges `items` according to the new `indices` that each item should occupy.

    This function is best illustrated by the example below.
    It also has an inverse available in 'unarrange_by_indices'.

    Example:
        >>> arrange_by_indices(['a', 'b', 'c'], [1, 2, 0])
        ['b', 'c', 'a']
    """
    return [items[i] for i in indices]


def unarrange_by_indices(items, indices):
    """The inverse function to 'arrange_by_indices'.

    Ex:
    unarrange_by_indices(['b', 'c', 'a'], [1, 2, 0])
     => ['a', 'b', 'c']
    """
    return [x for x, _ in sorted(zip(indices, items), key=itemgetter(0))]


ATTRS_MAP = {
    "PuPol": "pump_pol",
    "PrPol": "probe_pol",
    "SFLNM0": "lens_mode",
    "Lens Mode": "lens_mode",
    "Excitation Energy": "hv",
    "SFPE_0": "pass_energy",
    "Pass Energy": "pass_energy",
    "Slit Plate": "slit",
    "Number of Sweeps": "n_sweeps",
    "Acquisition Mode": "scan_mode",
    "Region Name": "scan_region",
    "Instrument": "instrument",
    "Pressure": "pressure",
    "User": "user",
    "Polar": "theta",
    "Theta": "theta",
    "Sample": "sample",
    "Beta": "beta",
    "Azimuth": "chi",
    "Location": "location",
}



"""Allows for making any function of a spectrum into a dynamic tool with Bokeh."""
import inspect

import numpy as np

import typing
from arpes.exceptions import AnalysisError
from arpes.plotting.interactive_utils import BokehInteractiveTool, CursorTool
from arpes.typing import DataType
from arpes.utilities import Debounce, normalize_to_spectrum

__all__ = (
    "DynamicTool",
    "dyn",
)


class DynamicTool(BokehInteractiveTool, CursorTool):
    """Presents a utility to rerun a function with different arguments and see the result of the function."""

    def __init__(self, analysis_fn, widget_specification, **kwargs):
        """Initialize the tool and load settings from the user specified ones."""
        super().__init__()

        self.load_settings(**kwargs)
        self.analysis_fn = analysis_fn
        self.widget_specification = widget_specification

        self.app_main_size = self.settings.get("main_width", 600)
        self.app_marginal_size = self.settings.get("marginal_width", 300)

    def tool_handler(self, doc):
        """Configures widgets for the dynamic tool.

        In order to accomplish this, we need to inspect the type signature
        for the function and generate inputs for it dynamically.
        """
        from bokeh import events
        from bokeh.layouts import row, column
        from bokeh.models.mappers import LinearColorMapper
        from bokeh.models import widgets
        from bokeh.plotting import figure

        if len(self.arr.shape) != 2:
            raise AnalysisError("Cannot use the band tool on non image-like spectra")

        self.data_for_display = self.arr
        x_coords, y_coords = (
            self.data_for_display.coords[self.data_for_display.dims[0]],
            self.data_for_display.coords[self.data_for_display.dims[1]],
        )

        default_palette = self.default_palette

        self.app_context.update(
            {
                "data": self.arr,
                "data_range": {
                    "x": (np.min(x_coords.values), np.max(x_coords.values)),
                    "y": (np.min(y_coords.values), np.max(y_coords.values)),
                },
            }
        )

        figures, plots = self.app_context["figures"], self.app_context["plots"]

        self.cursor = [np.mean(self.data_range["x"]), np.mean(self.data_range["y"])]

        self.color_maps["main"] = LinearColorMapper(
            default_palette,
            low=np.min(self.data_for_display.values),
            high=np.max(self.data_for_display.values),
            nan_color="black",
        )

        main_tools = ["wheel_zoom", "tap", "reset"]
        main_title = "{} Tool: WARNING Unidentified".format(self.analysis_fn.__name__)

        try:
            main_title = "{} Tool: {}".format(
                self.analysis_fn.__name__, self.data_for_display.S.label[:60]
            )
        except:
            pass

        figures["main"] = figure(
            tools=main_tools,
            plot_width=self.app_main_size,
            plot_height=self.app_main_size,
            min_border=10,
            min_border_left=50,
            toolbar_location="left",
            x_axis_location="below",
            y_axis_location="right",
            title=main_title,
            x_range=self.data_range["x"],
            y_range=self.data_range["y"],
        )
        figures["main"].xaxis.axis_label = self.data_for_display.dims[0]
        figures["main"].yaxis.axis_label = self.data_for_display.dims[1]
        figures["main"].toolbar.logo = None
        figures["main"].background_fill_color = "#fafafa"
        plots["main"] = figures["main"].image(
            [self.data_for_display.values.T],
            x=self.app_context["data_range"]["x"][0],
            y=self.app_context["data_range"]["y"][0],
            dw=self.app_context["data_range"]["x"][1] - self.app_context["data_range"]["x"][0],
            dh=self.app_context["data_range"]["y"][1] - self.app_context["data_range"]["y"][0],
            color_mapper=self.app_context["color_maps"]["main"],
        )

        # Create the bottom marginal plot
        bottom_marginal = self.data_for_display.sel(
            **dict([[self.data_for_display.dims[1], self.cursor[1]]]), method="nearest"
        )
        bottom_marginal_original = self.arr.sel(
            **dict([[self.data_for_display.dims[1], self.cursor[1]]]), method="nearest"
        )
        figures["bottom_marginal"] = figure(
            plot_width=self.app_main_size,
            plot_height=200,
            title=None,
            x_range=figures["main"].x_range,
            y_range=(np.min(bottom_marginal.values), np.max(bottom_marginal.values)),
            x_axis_location="above",
            toolbar_location=None,
            tools=[],
        )
        plots["bottom_marginal"] = figures["bottom_marginal"].line(
            x=bottom_marginal.coords[self.data_for_display.dims[0]].values, y=bottom_marginal.values
        )
        plots["bottom_marginal_original"] = figures["bottom_marginal"].line(
            x=bottom_marginal_original.coords[self.arr.dims[0]].values,
            y=bottom_marginal_original.values,
            line_color="red",
        )

        # Create the right marginal plot
        right_marginal = self.data_for_display.sel(
            **dict([[self.data_for_display.dims[0], self.cursor[0]]]), method="nearest"
        )
        right_marginal_original = self.arr.sel(
            **dict([[self.data_for_display.dims[0], self.cursor[0]]]), method="nearest"
        )
        figures["right_marginal"] = figure(
            plot_width=200,
            plot_height=self.app_main_size,
            title=None,
            y_range=figures["main"].y_range,
            x_range=(np.min(right_marginal.values), np.max(right_marginal.values)),
            y_axis_location="left",
            toolbar_location=None,
            tools=[],
        )
        plots["right_marginal"] = figures["right_marginal"].line(
            y=right_marginal.coords[self.data_for_display.dims[1]].values, x=right_marginal.values
        )
        plots["right_marginal_original"] = figures["right_marginal"].line(
            y=right_marginal_original.coords[self.data_for_display.dims[1]].values,
            x=right_marginal_original.values,
            line_color="red",
        )

        # add lines
        self.add_cursor_lines(figures["main"])
        _ = figures["main"].multi_line(xs=[], ys=[], line_color="white", line_width=1)  # band lines

        # prep the widgets for the analysis function
        signature = inspect.signature(self.analysis_fn)

        # drop the first which has to be the input data, we can revisit later if this is too limiting
        parameter_names = list(signature.parameters)[1:]
        named_widgets = dict(zip(parameter_names, self.widget_specification))
        built_widgets = {}

        def update_marginals():
            right_marginal_data = self.data_for_display.sel(
                **dict([[self.data_for_display.dims[0], self.cursor[0]]]), method="nearest"
            )
            bottom_marginal_data = self.data_for_display.sel(
                **dict([[self.data_for_display.dims[1], self.cursor[1]]]), method="nearest"
            )
            plots["bottom_marginal"].data_source.data = {
                "x": bottom_marginal_data.coords[self.data_for_display.dims[0]].values,
                "y": bottom_marginal_data.values,
            }
            plots["right_marginal"].data_source.data = {
                "y": right_marginal_data.coords[self.data_for_display.dims[1]].values,
                "x": right_marginal_data.values,
            }

            right_marginal_data = self.arr.sel(
                **dict([[self.data_for_display.dims[0], self.cursor[0]]]), method="nearest"
            )
            bottom_marginal_data = self.arr.sel(
                **dict([[self.data_for_display.dims[1], self.cursor[1]]]), method="nearest"
            )
            plots["bottom_marginal_original"].data_source.data = {
                "x": bottom_marginal_data.coords[self.data_for_display.dims[0]].values,
                "y": bottom_marginal_data.values,
            }
            plots["right_marginal_original"].data_source.data = {
                "y": right_marginal_data.coords[self.data_for_display.dims[1]].values,
                "x": right_marginal_data.values,
            }
            figures["bottom_marginal"].y_range.start = np.min(bottom_marginal_data.values)
            figures["bottom_marginal"].y_range.end = np.max(bottom_marginal_data.values)
            figures["right_marginal"].x_range.start = np.min(right_marginal_data.values)
            figures["right_marginal"].x_range.end = np.max(right_marginal_data.values)

        def click_main_image(event):
            self.cursor = [event.x, event.y]
            update_marginals()

        error_msg = widgets.Div(text="")

        @Debounce(0.25)
        def update_data_for_display():
            try:
                self.data_for_display = self.analysis_fn(
                    self.arr,
                    *[built_widgets[p].value for p in parameter_names if p in built_widgets]
                )
                error_msg.text = ""
            except Exception as e:
                error_msg.text = "{}".format(e)

            # flush + update
            update_marginals()
            plots["main"].data_source.data = {"image": [self.data_for_display.values.T]}

        def update_data_change_wrapper(attr, old, new):
            if old != new:
                update_data_for_display()

        for parameter_name in named_widgets.keys():
            specification = named_widgets[parameter_name]

            widget = None
            if specification["type"] == int:
                widget = widgets.Slider(
                    start=specification["start"],
                    end=specification["end"],
                    value=specification["value"],
                    title=parameter_name,
                )
            if specification["type"] == float:
                widget = widgets.Slider(
                    start=specification["start"],
                    end=specification["end"],
                    value=specification["value"],
                    step=specification["step"],
                    title=parameter_name,
                )

            if widget is not None:
                built_widgets[parameter_name] = widget
                widget.on_change("value", update_data_change_wrapper)

        update_main_colormap = self.update_colormap_for("main")

        self.app_context["run"] = lambda x: x

        main_color_range_slider = widgets.RangeSlider(
            start=0,
            end=100,
            value=(
                0,
                100,
            ),
            title="Color Range",
        )

        # Attach callbacks
        main_color_range_slider.on_change("value", update_main_colormap)

        figures["main"].on_event(events.Tap, click_main_image)

        layout = row(
            column(figures["main"], figures["bottom_marginal"]),
            column(figures["right_marginal"]),
            column(
                column(*[built_widgets[p] for p in parameter_names if p in built_widgets]),
                column(
                    self._cursor_info,
                    main_color_range_slider,
                    error_msg,
                ),
            ),
        )

        doc.add_root(layout)
        doc.title = "Band Tool"


def dyn(dynamic_function: typing.Callable, data: DataType, widget_specifications=None):
    """Starts the dynamic tool using `dynamic_function` and widgets for each arg."""
    data = normalize_to_spectrum(data)

    tool = DynamicTool(dynamic_function, widget_specifications)
    return tool.make_tool(data)



"""Top level module for PyARPES."""

# pylint: disable=unused-import

import warnings

from typing import Union

# Use both version conventions for people's sanity.
VERSION = "3.0.1"
__version__ = VERSION


def check() -> None:
    """Verifies certain aspects of the installation and provides guidance broken installations."""

    def verify_qt_tool() -> Union[str, None]:
        pip_command = "pip install pyqtgraph"
        warning = (
            "Using qt_tool, the PyARPES version of Image Tool, requires "
            "pyqtgraph and Qt5:\n\n\tYou can install with: {}".format(pip_command)
        )
        try:
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                import pyqtgraph
        except ImportError:
            return warning
        return None

    # def verify_igor_pro() -> Union[str, None]:
    #     pip_command = "pip install https://github.com/chstan/igorpy/tarball/712a4c4#egg=igor-0.3.1"
    #     warning = "For Igor support, install igorpy with: {}".format(pip_command)
    #     warning_incompatible = (
    #         "PyARPES requires a patched copy of igorpy, "
    #         "available at \n\thttps://github.com/chstan/igorpy/tarball/712a4c4\n\n\tYou can install with: "
    #         "{}".format(pip_command)
    #     )
    #     try:
    #         import igor

    #         if igor.__version__ <= "0.3":
    #             raise ValueError("Not using patched version of igorpy.")

    #     except ValueError:
    #         return warning_incompatible
    #     except (ImportError, AttributeError):
    #         return warning

    #     return None

    def verify_bokeh() -> Union[str, None]:
        pip_command = "pip install bokeh>=2.0.0,<3.0.0"

        warning = "For bokeh support, install version 2.3.x\n\t with {}".format(
            pip_command
        )
        warning_incompatible = (
            "PyARPES, requires version 2 of bokeh. You can install with \n\t{}".format(
                pip_command
            )
        )

        try:
            import bokeh

            if not bokeh.__version__.startswith("2."):
                raise ValueError("Not using the specified version of Bokeh.")

        except ImportError:
            return warning
        except ValueError:
            return warning_incompatible
        return None

    checks = [
        # ("Igor Pro Support", verify_igor_pro),
        ("Bokeh Support", verify_bokeh),
        ("qt_tool Support", verify_qt_tool),
    ]

    from colorama import Fore, Style

    print("Checking...")
    for check_name, check_fn in checks:
        initial_str = "[ ] {}".format(check_name)
        print(initial_str, end="", flush=True)

        failure_message = check_fn()

        print(
            "\b" * len(initial_str) + " " * len(initial_str) + "\b" * len(initial_str),
            end="",
        )

        if failure_message is None:
            print("{}[✔] {}{}".format(Fore.GREEN, check_name, Style.RESET_ALL))
        else:
            print(
                "{}[✘] {}: \n\t{}{}".format(
                    Fore.RED, check_name, failure_message, Style.RESET_ALL
                )
            )



"""Definitions of common backgrounds."""

from lmfit.models import update_param_vals
import numpy as np

from .x_model_mixin import XModelMixin
from .functional_forms import affine_bkg

__all__ = ["AffineBackgroundModel"]


class AffineBackgroundModel(XModelMixin):
    """A model for an affine background."""

    def __init__(self, independent_vars=("x",), prefix="", missing="raise", name=None, **kwargs):
        """Defer to lmfit for initialization."""
        kwargs.update({"prefix": prefix, "missing": missing, "independent_vars": independent_vars})
        super().__init__(affine_bkg, **kwargs)

    def guess(self, data, x=None, **kwargs):
        """Use the tenth percentile value for the slope and a zero offset.

        Generally this should converge well regardless.
        """
        pars = self.make_params()

        pars["%slin_bkg" % self.prefix].set(value=np.percentile(data, 10))
        pars["%sconst_bkg" % self.prefix].set(value=0)

        return update_param_vals(pars, self.prefix, **kwargs)



"""Provides a Qt based implementation of Igor's ImageTool."""

from .base import ImageTool
from .parent_tool import ParentTool
from arpes.typing import DataType
from arpes.utilities.qt import run_tool_in_daemon_process


def _image_tool(data: DataType, **kwargs):
    """Starts the image_tool using an input spectrum."""

    tool = ImageTool()
    tool.set_data(data)
    tool.start(**kwargs)


def _parent_tool(data: DataType, children: list[tuple[DataType, str]], **kwargs):
    """Starts the parent_tool using an input spectrum."""

    tool = ParentTool()
    tool.set_data(data)
    for child in children:
        tool.add_child(*child)
    tool.start(**kwargs)


image_tool = run_tool_in_daemon_process(_image_tool)
parent_tool = run_tool_in_daemon_process(_parent_tool)



"""Plotting utilities related to density of states plots."""
import matplotlib
import numpy as np
from matplotlib import pyplot as plt
from matplotlib import colors, gridspec

from arpes.analysis.xps import approximate_core_levels
from arpes.provenance import save_plot_provenance
from arpes.utilities import normalize_to_spectrum

from .utils import path_for_plot, savefig

__all__ = (
    "plot_dos",
    "plot_core_levels",
)


@save_plot_provenance
def plot_core_levels(
    data,
    title=None,
    out=None,
    norm=None,
    dos_pow=1,
    core_levels=None,
    binning=1,
    promenance=5,
    **kwargs
):
    """Plots an XPS curve and approximate core level locations."""
    axes, cbar = plot_dos(data=data, title=title, out=None, norm=norm, dos_pow=dos_pow, **kwargs)

    if core_levels is None:
        core_levels = approximate_core_levels(data, binning=binning, promenance=promenance)

    for core_level in core_levels:
        axes[1].axvline(core_level, color="red", ls="--")

    if out is not None:
        savefig(out, dpi=400)
        return path_for_plot(out)
    else:
        return axes, cbar


@save_plot_provenance
def plot_dos(data, title=None, out=None, norm=None, dos_pow=1, **kwargs):
    """Plots the density of states (momentum integrated) image next to the original spectrum."""
    data = normalize_to_spectrum(data)

    fig = plt.figure(figsize=(14, 6))
    fig.subplots_adjust(hspace=0.00)
    gs = gridspec.GridSpec(2, 1, height_ratios=[3, 1])

    ax0 = plt.subplot(gs[0])
    axes = (ax0, plt.subplot(gs[1], sharex=ax0))

    data.values[np.isnan(data.values)] = 0
    cbar_axes = matplotlib.colorbar.make_axes(axes, pad=0.01)

    mesh = data.plot(ax=axes[0], norm=norm or colors.PowerNorm(gamma=0.15))

    axes[1].set_facecolor((0.95, 0.95, 0.95))
    density_of_states = data.S.sum_other(["eV"])
    (density_of_states ** dos_pow).plot(ax=axes[1])

    cbar = plt.colorbar(mesh, cax=cbar_axes[0])
    cbar.set_label("Photoemission Intensity (Arb.)")

    axes[1].set_ylabel("Spectrum DOS", labelpad=12)
    axes[0].set_title(title or "")

    if out is not None:
        savefig(out, dpi=400)
        return path_for_plot(out)
    else:
        return fig, axes, cbar



"""Provides extremely fast 2D and 3D linear interpolation.

This is used for momentum conversion in place of the scipy
GridInterpolator where it is possible to do so. It is many many 
times faster than the grid interpolator and together with other optimizations
resulted in a 50x improvement in the momentum conversion time for
ARPES data in PyARPES.
"""
import numba
from dataclasses import dataclass
from typing import List, Union
import math
import numpy as np

__all__ = [
    "Interpolator",
]


@numba.njit
def to_fractional_coordinate(coord, initial, delta):
    return (coord - initial) / delta


@numba.njit
def _i1d(xd, c0, c1):
    return c0 * (1 - xd) + c1 * xd


@numba.njit
def raw_lin_interpolate_1d(xd, c0, c1):
    return _i1d(xd, c0, c1)


@numba.njit
def raw_lin_interpolate_2d(xd, yd, c00, c01, c10, c11):
    # project to 1D
    c0 = _i1d(xd, c00, c10)
    c1 = _i1d(xd, c01, c11)

    return _i1d(yd, c0, c1)


@numba.njit
def raw_lin_interpolate_3d(xd, yd, zd, c000, c001, c010, c100, c011, c101, c110, c111):
    # project to 2D
    c00 = _i1d(xd, c000, c100)
    c01 = _i1d(xd, c001, c101)
    c10 = _i1d(xd, c010, c110)
    c11 = _i1d(xd, c011, c111)

    # project to 1D
    c0 = _i1d(yd, c00, c10)
    c1 = _i1d(yd, c01, c11)

    return _i1d(zd, c0, c1)


@numba.njit
def lin_interpolate_3d(data, ix, iy, iz, ixp, iyp, izp, xd, yd, zd):
    return raw_lin_interpolate_3d(
        xd,
        yd,
        zd,
        data[ix][iy][iz],
        data[ix][iy][izp],
        data[ix][iyp][iz],
        data[ixp][iy][iz],
        data[ix][iyp][izp],
        data[ixp][iy][izp],
        data[ixp][iyp][iz],
        data[ixp][iyp][izp],
    )


@numba.njit
def lin_interpolate_2d(data, ix, iy, ixp, iyp, xd, yd):
    return raw_lin_interpolate_2d(
        xd,
        yd,
        data[ix][iy],
        data[ix][iyp],
        data[ixp][iy],
        data[ixp][iyp],
    )


@numba.njit(parallel=True)
def interpolate_3d(
    data,
    output,
    lower_corner_x,
    lower_corner_y,
    lower_corner_z,
    delta_x,
    delta_y,
    delta_z,
    shape_x,
    shape_y,
    shape_z,
    x,
    y,
    z,
    fill_value=np.nan,
):
    for i in numba.prange(len(x)):
        if np.isnan(x[i]) or np.isnan(y[i]) or np.isnan(z[i]):
            output[i] = fill_value
            continue

        ix = to_fractional_coordinate(x[i], lower_corner_x, delta_x)
        iy = to_fractional_coordinate(y[i], lower_corner_y, delta_y)
        iz = to_fractional_coordinate(z[i], lower_corner_z, delta_z)

        if ix < 0 or iy < 0 or iz < 0 or ix >= shape_x or iy >= shape_y or iz >= shape_z:
            output[i] = fill_value
            continue

        iix, iiy, iiz = math.floor(ix), math.floor(iy), math.floor(iz)
        iixp, iiyp, iizp = (
            min(iix + 1, shape_x - 1),
            min(iiy + 1, shape_y - 1),
            min(iiz + 1, shape_z - 1),
        )
        xd, yd, zd = ix - iix, iy - iiy, iz - iiz

        output[i] = lin_interpolate_3d(data, iix, iiy, iiz, iixp, iiyp, iizp, xd, yd, zd)


@numba.njit(parallel=True)
def interpolate_2d(
    data,
    output,
    lower_corner_x,
    lower_corner_y,
    delta_x,
    delta_y,
    shape_x,
    shape_y,
    x,
    y,
    fill_value=np.nan,
):
    for i in numba.prange(len(x)):
        if np.isnan(x[i]) or np.isnan(y[i]):
            output[i] = fill_value
            continue

        ix = to_fractional_coordinate(x[i], lower_corner_x, delta_x)
        iy = to_fractional_coordinate(y[i], lower_corner_y, delta_y)

        if ix < 0 or iy < 0 or ix >= shape_x - 1 or iy >= shape_y - 1:
            output[i] = fill_value
            continue

        iix, iiy = math.floor(ix), math.floor(iy)
        iixp, iiyp = (
            min(iix + 1, shape_x - 1),
            min(iiy + 1, shape_y - 1),
        )
        xd, yd = ix - iix, iy - iiy

        output[i] = lin_interpolate_2d(data, iix, iiy, iixp, iiyp, xd, yd)


@dataclass
class Interpolator:
    """Provides a Pythonic interface to fast gridded linear interpolation.

    More or less a drop-in replacement for scipy's RegularGridInterpolator,
    but much faster at the expense of not supporting any extrapolation.
    """

    lower_corner: List[float]
    delta: List[float]
    shape: List[int]
    data: np.ndarray

    def __post_init__(self):
        """Convert data to floating point representation.

        Because we do linear not nearest neighbor interpolation this should be safe
        always.
        """
        self.data = self.data.astype(np.float64, copy=False)

    @classmethod
    def from_arrays(cls, xyz: List[np.ndarray], data: np.ndarray):
        """Initializes the interpreter from a coordinate and data array.

        Args:
            xyz: A list of the coordinate arrays. Should be length 2 or 3
              because we provide 2D and 3D coordinate interpolation.
            data: The value of the interpolated function at the coordinate in `xyz`
        """
        lower_corner = [xi[0] for xi in xyz]
        delta = [xi[1] - xi[0] for xi in xyz]
        shape = [len(xi) for xi in xyz]
        return cls(lower_corner, delta, shape, data)

    def __call__(self, xi: Union[np.ndarray, List[np.ndarray]]) -> np.ndarray:
        """Performs linear interpolation at the coordinates given by `xi`.

        Whether 2D or 3D interpolation is used depends on the dimensionality of `xi` and
        `self.data` but of course they must match one another.

        Args:
            xi: A list or stacked array of the coordinates. Provides a [d, k] array
              of k points each with d dimensions/indices.

        Returns:
            The interpolated values f(x_i) at each point x_i, as a length k scalar array.
        """
        if isinstance(xi, np.ndarray):
            xi = xi.astype(np.float64, copy=False)
            xi = [xi[:, i] for i in range(self.data.ndim)]
        else:
            xi = [xii.astype(np.float64, copy=False) for xii in xi]

        output = np.zeros_like(xi[0])

        interpolator = {
            3: interpolate_3d,
            2: interpolate_2d,
        }[self.data.ndim]

        interpolator(
            self.data,
            output,
            *self.lower_corner,
            *self.delta,
            *self.shape,
            *xi,
        )

        return output



"""Specialized type annotations for use in PyARPES.

In particular, we frequently allow using the `DataType` annotation,
which refers to either an xarray.DataArray or xarray.Dataset.

Additionally, we often use `NormalizableDataType` which
means essentially anything that can be turned into a dataset,
for instance by loading from the cache using an ID, or which is
literally already data.
"""

import uuid

import xarray as xr
import numpy as np

__all__ = ["DTypeLike", "DataType", "NormalizableDataType", "xr_types"]

DTypeLike = np.dtype | None | type | str

DataType = xr.DataArray | xr.Dataset
NormalizableDataType = DataType | str | uuid.UUID

xr_types = (xr.DataArray, xr.Dataset)



"""Implements a 2D and 3D data browser via Bokeh."""
import copy
import warnings

import numpy as np

import colorcet as cc
from arpes.fits import ExponentialDecayCModel, GStepBModel

from .interactive_utils import CursorTool, SaveableTool

__all__ = ("ImageTool",)

# TODO Implement alignment tool


class ImageTool(SaveableTool, CursorTool):
    """Implements a 2D and 3D data browser via Bokeh."""

    def __init__(self, curs=None, **kwargs):
        """Load application and fetch marginal sizes from settings."""
        super().__init__(name=kwargs.pop("name", None))

        self.load_settings(**kwargs)

        self.app_main_size = self.settings.get("main_width", 600)
        self.app_marginal_size = self.settings.get("marginal_width", 300)

        if curs is not None:
            self.cursor_default = curs

    # TODO select path in image
    def prep_image(self, image_arr):
        """Optionally, postprocess data before showing it."""
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            if self.app_context["color_mode"] == "linear":
                return image_arr.values

            # avoid dependency conflict with numpy v0.16 for now
            from skimage import exposure  # pylint: disable=import-error

            return exposure.equalize_adapthist(image_arr.values, clip_limit=0.03)

    def tool_handler(self, doc):
        """Delegates widget creation to 2D and 3D code based on array dimensions."""
        if len(self.arr.shape) == 3:
            return self.tool_handler_3d(doc)

        return self.tool_handler_2d(doc)

    def tool_handler_2d(self, doc):
        """Application definition and widgets for the 2D data browser."""
        from bokeh import events
        from bokeh.layouts import row, column, Spacer
        from bokeh.models import ColumnDataSource, widgets
        from bokeh.models.mappers import LinearColorMapper
        from bokeh.models.widgets.markups import Div
        from bokeh.plotting import figure

        arr = self.arr
        # Set up the data
        x_coords, y_coords = arr.coords[arr.dims[0]], arr.coords[arr.dims[1]]

        # Styling
        default_palette = self.default_palette
        if arr.S.is_subtracted:
            default_palette = cc.coolwarm

        error_alpha = 0.3
        error_fill = "#3288bd"

        # Application Organization
        self.app_context.update(
            {
                "data": arr,
                "data_range": {
                    "x": (np.min(x_coords.values), np.max(x_coords.values)),
                    "y": (np.min(y_coords.values), np.max(y_coords.values)),
                },
                "show_stat_variation": False,
                "color_mode": "linear",
            }
        )

        def stats_patch_from_data(data, subsampling_rate=None):
            if subsampling_rate is None:
                subsampling_rate = int(min(data.values.shape[0] / 50, 5))
                if subsampling_rate == 0:
                    subsampling_rate = 1

            x_values = data.coords[data.dims[0]].values[::subsampling_rate]
            values = data.values[::subsampling_rate]
            sq = np.sqrt(values)
            lower, upper = values - sq, values + sq

            return {
                "x": np.append(x_values, x_values[::-1]),
                "y": np.append(lower, upper[::-1]),
            }

        def update_stat_variation(plot_name, data):
            patch_data = stats_patch_from_data(data)
            if plot_name != "right":  # the right plot is on transposed axes
                plots[plot_name + "_marginal_err"].data_source.data = patch_data
            else:
                plots[plot_name + "_marginal_err"].data_source.data = {
                    "x": patch_data["y"],
                    "y": patch_data["x"],
                }

        figures, plots, app_widgets = (
            self.app_context["figures"],
            self.app_context["plots"],
            self.app_context["widgets"],
        )

        if self.cursor_default is not None and len(self.cursor_default) == 2:
            self.cursor = self.cursor_default
        else:
            self.cursor = [
                np.mean(self.app_context["data_range"]["x"]),
                np.mean(self.app_context["data_range"]["y"]),
            ]  # try a sensible default

        # create the main inset plot
        main_image = arr
        prepped_main_image = self.prep_image(main_image)
        self.app_context["color_maps"]["main"] = LinearColorMapper(
            default_palette,
            low=np.min(prepped_main_image),
            high=np.max(prepped_main_image),
            nan_color="black",
        )

        main_tools = ["wheel_zoom", "tap", "reset", "save"]
        main_title = "Bokeh Tool: WARNING Unidentified"
        try:
            main_title = "Bokeh Tool: %s" % arr.S.label[:60]
        except:
            pass
        figures["main"] = figure(
            tools=main_tools,
            plot_width=self.app_main_size,
            plot_height=self.app_main_size,
            min_border=10,
            min_border_left=50,
            toolbar_location="left",
            x_axis_location="below",
            y_axis_location="right",
            title=main_title,
            x_range=self.app_context["data_range"]["x"],
            y_range=self.app_context["data_range"]["y"],
        )
        figures["main"].xaxis.axis_label = arr.dims[0]
        figures["main"].yaxis.axis_label = arr.dims[1]
        figures["main"].toolbar.logo = None
        figures["main"].background_fill_color = "#fafafa"
        plots["main"] = figures["main"].image(
            [prepped_main_image.T],
            x=self.app_context["data_range"]["x"][0],
            y=self.app_context["data_range"]["y"][0],
            dw=self.app_context["data_range"]["x"][1] - self.app_context["data_range"]["x"][0],
            dh=self.app_context["data_range"]["y"][1] - self.app_context["data_range"]["y"][0],
            color_mapper=self.app_context["color_maps"]["main"],
        )

        app_widgets["info_div"] = Div(text="", width=self.app_marginal_size, height=100)

        # Create the bottom marginal plot
        bottom_marginal = arr.sel(**dict([[arr.dims[1], self.cursor[1]]]), method="nearest")
        figures["bottom_marginal"] = figure(
            plot_width=self.app_main_size,
            plot_height=200,
            title=None,
            x_range=figures["main"].x_range,
            y_range=(np.min(bottom_marginal.values), np.max(bottom_marginal.values)),
            x_axis_location="above",
            toolbar_location=None,
            tools=[],
        )
        plots["bottom_marginal"] = figures["bottom_marginal"].line(
            x=bottom_marginal.coords[arr.dims[0]].values, y=bottom_marginal.values
        )
        plots["bottom_marginal_err"] = figures["bottom_marginal"].patch(
            x=[], y=[], color=error_fill, fill_alpha=error_alpha, line_color=None
        )

        # Create the right marginal plot
        right_marginal = arr.sel(**dict([[arr.dims[0], self.cursor[0]]]), method="nearest")
        figures["right_marginal"] = figure(
            plot_width=200,
            plot_height=self.app_main_size,
            title=None,
            y_range=figures["main"].y_range,
            x_range=(np.min(right_marginal.values), np.max(right_marginal.values)),
            y_axis_location="left",
            toolbar_location=None,
            tools=[],
        )
        plots["right_marginal"] = figures["right_marginal"].line(
            y=right_marginal.coords[arr.dims[1]].values, x=right_marginal.values
        )
        plots["right_marginal_err"] = figures["right_marginal"].patch(
            x=[], y=[], color=error_fill, fill_alpha=error_alpha, line_color=None
        )

        cursor_lines = self.add_cursor_lines(figures["main"])

        # Attach tools and callbacks
        toggle = widgets.Toggle(label="Show Stat. Variation", button_type="success", active=False)

        def set_show_stat_variation(should_show):
            self.app_context["show_stat_variation"] = should_show

            if should_show:
                main_image_data = arr
                update_stat_variation(
                    "bottom",
                    main_image_data.sel(**dict([[arr.dims[1], self.cursor[1]]]), method="nearest"),
                )
                update_stat_variation(
                    "right",
                    main_image_data.sel(**dict([[arr.dims[0], self.cursor[0]]]), method="nearest"),
                )
                plots["bottom_marginal_err"].visible = True
                plots["right_marginal_err"].visible = True
            else:
                plots["bottom_marginal_err"].visible = False
                plots["right_marginal_err"].visible = False

        toggle.on_click(set_show_stat_variation)

        scan_keys = ["x", "y", "z", "pass_energy", "hv", "location", "id", "probe_pol", "pump_pol"]
        scan_info_source = ColumnDataSource(
            {
                "keys": [k for k in scan_keys if k in arr.attrs],
                "values": [
                    str(v) if isinstance(v, float) and np.isnan(v) else v
                    for v in [arr.attrs[k] for k in scan_keys if k in arr.attrs]
                ],
            }
        )
        scan_info_columns = [
            widgets.TableColumn(field="keys", title="Attr."),
            widgets.TableColumn(field="values", title="Value"),
        ]

        POINTER_MODES = [
            ("Cursor", "cursor"),
            ("Path", "path"),
        ]

        COLOR_MODES = [
            ("Adaptive Hist. Eq. (Slow)", "adaptive_equalization"),
            # ('Histogram Eq.', 'equalization',), # not implemented
            ("Linear", "linear"),
            # ('Log', 'log',), # not implemented
        ]

        def on_change_color_mode(event):
            new_color_mode = event.item
            self.app_context["color_mode"] = new_color_mode
            if old is None or old != new_color_mode:
                right_image_data = arr.sel(
                    **dict([[arr.dims[0], self.cursor[0]]]), method="nearest"
                )
                bottom_image_data = arr.sel(
                    **dict([[arr.dims[1], self.cursor[1]]]), method="nearest"
                )
                main_image_data = arr
                prepped_right_image = self.prep_image(right_image_data)
                prepped_bottom_image = self.prep_image(bottom_image_data)
                prepped_main_image = self.prep_image(main_image_data)
                plots["right"].data_source.data = {"image": [prepped_right_image]}
                plots["bottom"].data_source.data = {"image": [prepped_bottom_image.T]}
                plots["main"].data_source.data = {"image": [prepped_main_image.T]}
                update_main_colormap(None, None, main_color_range_slider.value)

        color_mode_dropdown = widgets.Dropdown(
            label="Color Mode", button_type="primary", menu=COLOR_MODES
        )
        color_mode_dropdown.on_click(on_change_color_mode)

        symmetry_point_name_input = widgets.TextInput(title="Symmetry Point Name", value="G")
        snap_checkbox = widgets.CheckboxButtonGroup(labels=["Snap Axes"], active=[])
        place_symmetry_point_at_cursor_button = widgets.Button(
            label="Place Point", button_type="primary"
        )

        def update_symmetry_points_for_display():
            pass

        def place_symmetry_point():
            cursor_dict = dict(zip(arr.dims, self.cursor))
            skip_dimensions = {"eV", "delay", "cycle"}
            if "symmetry_points" not in arr.attrs:
                arr.attrs["symmetry_points"] = {}

            snap_distance = {
                "phi": 2,
                "beta": 2,
                "kx": 0.01,
                "ky": 0.01,
                "kz": 0.01,
                "kp": 0.01,
                "hv": 4,
            }

            cursor_dict = {k: v for k, v in cursor_dict.items() if k not in skip_dimensions}
            snapped = copy.copy(cursor_dict)

            if "Snap Axes" in [snap_checkbox.labels[i] for i in snap_checkbox.active]:
                for axis, value in cursor_dict.items():
                    options = [
                        point[axis]
                        for point in arr.attrs["symmetry_points"].values()
                        if axis in point
                    ]
                    options = sorted(options, key=lambda x: np.abs(x - value))
                    if options and np.abs(options[0] - value) < snap_distance[axis]:
                        snapped[axis] = options[0]

            arr.attrs["symmetry_points"][symmetry_point_name_input.value] = snapped

        place_symmetry_point_at_cursor_button.on_click(place_symmetry_point)

        main_color_range_slider = widgets.RangeSlider(
            start=0,
            end=100,
            value=(
                0,
                100,
            ),
            title="Color Range (Main)",
        )

        layout = row(
            column(figures["main"], figures["bottom_marginal"]),
            column(figures["right_marginal"], Spacer(width=200, height=200)),
            column(
                column(
                    widgets.Dropdown(
                        label="Pointer Mode", button_type="primary", menu=POINTER_MODES
                    )
                ),
                widgets.Tabs(
                    tabs=[
                        widgets.Panel(
                            child=column(
                                Div(text="<h2>Colorscale:</h2>"),
                                color_mode_dropdown,
                                main_color_range_slider,
                                Div(text='<h2 style="padding-top: 30px;">General Settings:</h2>'),
                                toggle,
                                self._cursor_info,
                                sizing_mode="scale_width",
                            ),
                            title="Settings",
                        ),
                        widgets.Panel(
                            child=column(
                                app_widgets["info_div"],
                                Div(
                                    text='<h2 style="padding-top: 30px; padding-bottom: 10px;">Scan Info</h2>'
                                ),
                                widgets.DataTable(
                                    source=scan_info_source,
                                    columns=scan_info_columns,
                                    width=400,
                                    height=400,
                                ),
                                sizing_mode="scale_width",
                                width=400,
                            ),
                            title="Info",
                        ),
                        widgets.Panel(
                            child=column(
                                Div(text="<h2>Preparation</h2>"),
                                symmetry_point_name_input,
                                snap_checkbox,
                                place_symmetry_point_at_cursor_button,
                                sizing_mode="scale_width",
                            ),
                            title="Preparation",
                        ),
                    ],
                    width=400,
                ),
            ),
        )

        update_main_colormap = self.update_colormap_for("main")

        def click_main_image(event):
            self.cursor = [event.x, event.y]

            right_marginal_data = arr.sel(**dict([[arr.dims[0], self.cursor[0]]]), method="nearest")
            bottom_marginal_data = arr.sel(
                **dict([[arr.dims[1], self.cursor[1]]]), method="nearest"
            )
            plots["bottom_marginal"].data_source.data = {
                "x": bottom_marginal_data.coords[arr.dims[0]].values,
                "y": bottom_marginal_data.values,
            }
            plots["right_marginal"].data_source.data = {
                "y": right_marginal_data.coords[arr.dims[1]].values,
                "x": right_marginal_data.values,
            }
            if self.app_context["show_stat_variation"]:
                update_stat_variation("right", right_marginal_data)
                update_stat_variation("bottom", bottom_marginal_data)
            figures["bottom_marginal"].y_range.start = np.min(bottom_marginal_data.values)
            figures["bottom_marginal"].y_range.end = np.max(bottom_marginal_data.values)
            figures["right_marginal"].x_range.start = np.min(right_marginal_data.values)
            figures["right_marginal"].x_range.end = np.max(right_marginal_data.values)

            self.save_app()

        figures["main"].on_event(events.Tap, click_main_image)
        main_color_range_slider.on_change("value", update_main_colormap)

        doc.add_root(layout)
        doc.title = "Bokeh Tool"
        self.load_app()
        self.save_app()

    def serialize(self):
        """Saves the current cursor position so it can be loaded later."""
        return {
            "cursor_dict": self.cursor_dict,
            "cursor": self.cursor,
        }

    def deserialize(self, json_data):
        """Loads the cursor position from a saved copy of the tool."""
        if "cursor" in json_data:
            self.cursor = json_data["cursor"]

    def tool_handler_3d(self, doc):
        """Application and widget definitions for the 3D data browser."""
        from bokeh import events
        from bokeh.layouts import row, column, Spacer
        from bokeh.models import ColumnDataSource, HoverTool, widgets
        from bokeh.models.mappers import LinearColorMapper
        from bokeh.models.widgets.markups import Div
        from bokeh.plotting import figure

        arr = self.arr
        # Set up the data
        x_coords, y_coords, z_coords = (
            arr.coords[arr.dims[0]],
            arr.coords[arr.dims[1]],
            arr.coords[arr.dims[2]],
        )

        info_formatters = {
            "eV": """<div>
                            <h2>Fermi Edge Info:</h2>
                            <p>Gap: <b>{:.1f} meV</b></p>
                            <p>Edge Width: <b>{:.1f} meV</b></p>
                            </div>
                            """,
            "delay": """<div>
                            <h2>Delay Scan Info:</h2>
                            <p>t0: <b>{:.1f} fs</b></p>
                            <p>Decay time: <b>{:.1f} fs</b></p>
                            </div>
                            """,
        }

        # Styling
        default_palette = self.default_palette
        if arr.S.is_subtracted:
            default_palette = cc.coolwarm
        error_alpha = 0.3
        error_fill = "#3288bd"

        # Application Organization
        self.app_context.update(
            {
                "data": arr,
                "data_range": {
                    "x": (np.min(x_coords.values), np.max(x_coords.values)),
                    "y": (np.min(y_coords.values), np.max(y_coords.values)),
                    "z": (np.min(z_coords.values), np.max(z_coords.values)),
                },
                "show_stat_variation": False,
                "color_mode": "linear",
            }
        )

        def stats_patch_from_data(data, subsampling_rate=None):
            if subsampling_rate is None:
                subsampling_rate = int(min(data.values.shape[0] / 50, 5))
                if subsampling_rate == 0:
                    subsampling_rate = 1

            x_values = data.coords[data.dims[0]].values[::subsampling_rate]
            values = data.values[::subsampling_rate]
            sq = np.sqrt(values)
            lower, upper = values - sq, values + sq

            return {
                "x": np.append(x_values, x_values[::-1]),
                "y": np.append(lower, upper[::-1]),
            }

        def update_stat_variation(plot_name, data):
            patch_data = stats_patch_from_data(data)
            if plot_name != "right":  # the right plot is on transposed axes
                plots[plot_name + "_marginal_err"].data_source.data = patch_data
            else:
                plots[plot_name + "_marginal_err"].data_source.data = {
                    "x": patch_data["y"],
                    "y": patch_data["x"],
                }

        figures, plots, app_widgets = (
            self.app_context["figures"],
            self.app_context["plots"],
            self.app_context["widgets"],
        )

        if self.cursor_default is not None and len(self.cursor_default) == 3:
            self.cursor = self.cursor_default
        else:
            self.cursor = [
                np.mean(self.data_range["x"]),
                np.mean(self.data_range["y"]),
                np.mean(self.data_range["z"]),
            ]  # try a sensible default

        # create the main inset plot
        main_image = arr.sel(**dict([[arr.dims[2], self.cursor[2]]]), method="nearest")
        prepped_main_image = self.prep_image(main_image)
        self.app_context["color_maps"]["main"] = LinearColorMapper(
            default_palette,
            low=np.min(prepped_main_image),
            high=np.max(prepped_main_image),
            nan_color="black",
        )

        main_title = "Bokeh Tool: WARNING Unidentified"
        try:
            main_title = "Bokeh Tool: %s" % arr.S.label[:60]
        except:
            pass

        main_tools = ["wheel_zoom", "tap", "reset", "save"]
        figures["main"] = figure(
            tools=main_tools,
            plot_width=self.app_main_size,
            plot_height=self.app_main_size,
            min_border=10,
            min_border_left=50,
            toolbar_location="left",
            x_axis_location="below",
            y_axis_location="right",
            title=main_title,
            x_range=self.app_context["data_range"]["x"],
            y_range=self.app_context["data_range"]["y"],
        )
        figures["main"].xaxis.axis_label = arr.dims[0]
        figures["main"].yaxis.axis_label = arr.dims[1]
        figures["main"].toolbar.logo = None
        figures["main"].background_fill_color = "#fafafa"
        plots["main"] = figures["main"].image(
            [prepped_main_image.T],
            x=self.app_context["data_range"]["x"][0],
            y=self.app_context["data_range"]["y"][0],
            dw=self.app_context["data_range"]["x"][1] - self.app_context["data_range"]["x"][0],
            dh=self.app_context["data_range"]["y"][1] - self.app_context["data_range"]["y"][0],
            color_mapper=self.app_context["color_maps"]["main"],
        )

        # Create the z-selector
        z_marginal_data = arr.sel(
            **dict([[arr.dims[0], self.cursor[0]], [arr.dims[1], self.cursor[1]]]), method="nearest"
        )
        z_hover_tool = HoverTool(
            tooltips=[
                ("x", "@x{1.111}"),
                ("Int. (arb.)", "@y{1.1}"),
            ],
            mode="vline",
        )
        z_tools = [z_hover_tool, "wheel_zoom", "tap", "reset"]
        figures["z_marginal"] = figure(
            plot_width=self.app_marginal_size,
            plot_height=self.app_marginal_size,
            x_range=self.app_context["data_range"]["z"],
            y_range=(np.min(z_marginal_data.values), np.max(z_marginal_data.values)),
            x_axis_location="above",
            toolbar_location="below",
            y_axis_location="right",
            tools=z_tools,
        )
        figures["z_marginal"].xaxis.major_label_text_font_size = "0pt"
        figures["z_marginal"].xaxis.axis_label = arr.dims[2]
        figures["z_marginal"].toolbar.logo = None
        plots["z_marginal"] = figures["z_marginal"].line(
            x=z_coords.values, y=z_marginal_data.values
        )
        plots["z_marginal_err"] = figures["z_marginal"].patch(
            x=[], y=[], color=error_fill, fill_alpha=error_alpha, line_color=None
        )

        info_formatter = info_formatters.get(arr.dims[2], "")
        app_widgets["info_div"] = Div(text="", width=self.app_marginal_size, height=100)

        if arr.dims[2] == "eV":
            # Try to fit a Fermi edge and display it in the plot
            try:
                fit_data = z_marginal_data.sel(eV=slice(-0.25, 0.2))
                z_fit = GStepBModel().guess_fit(fit_data)
                plots["z_fit"] = figures["z_marginal"].line(
                    x=fit_data.coords["eV"].values,
                    y=z_fit.best_fit,
                    line_dash="dashed",
                    line_color="red",
                )
                self.app_context["z_model"] = GStepBModel

                app_widgets["info_div"].text = info_formatter.format(
                    z_fit.params["center"].value * 1000,
                    z_fit.params["width"].value * 1000,
                )
            except:
                pass

        elif arr.dims[2] == "delay" and "t0" in arr.attrs:
            # Try to fit a decay constant to the data after t0
            plots["t0_marker"] = figures["z_marginal"].line(
                x=[float(arr.attrs.get("t0")), float(arr.attrs.get("t0"))],
                y=[0, 1000000],
                line_color="black",
                line_dash="dashed",
            )
            t0 = float(arr.attrs["t0"])
            self.app_context["z_model"] = ExponentialDecayCModel
            try:
                after_t0 = z_marginal_data.sel(delay=slice(t0 - 0.2, None))
                exp_model = ExponentialDecayCModel()
                z_fit = exp_model.guess_fit(after_t0, params={"t0": {"value": t0}})
                plots["z_fit"] = figures["z_marginal"].line(
                    x=after_t0.coords["delay"].values,
                    y=z_fit.best_fit,
                    line_dash="dashed",
                    line_color="red",
                )
            except Exception as e:
                plots["z_fit"] = figures["z_marginal"].line(
                    x=[], y=[], line_dash="dashed", line_color="red"
                )

        # Create the bottom marginal plot
        bottom_image = arr.sel(**dict([[arr.dims[1], self.cursor[1]]]), method="nearest")
        prepped_bottom_image = self.prep_image(bottom_image)
        self.app_context["color_maps"]["bottom"] = LinearColorMapper(
            default_palette,
            low=np.min(prepped_bottom_image),
            high=np.max(prepped_bottom_image),
            nan_color="black",
        )
        figures["bottom"] = figure(
            plot_width=self.app_main_size,
            plot_height=self.app_marginal_size,
            title=None,
            x_range=figures["main"].x_range,
            y_range=figures["z_marginal"].x_range,
            x_axis_location="above",
            toolbar_location=None,
            tools=[],
        )
        figures["bottom"].xaxis.major_label_text_font_size = "0pt"
        plots["bottom"] = figures["bottom"].image(
            [prepped_bottom_image.T],
            x=self.app_context["data_range"]["x"][0],
            y=self.app_context["data_range"]["z"][0],
            dw=self.app_context["data_range"]["x"][1] - self.app_context["data_range"]["x"][0],
            dh=self.app_context["data_range"]["z"][1] - self.app_context["data_range"]["z"][0],
            color_mapper=self.app_context["color_maps"]["bottom"],
        )
        bottom_marginal = bottom_image.sel(
            **dict([[arr.dims[2], self.cursor[2]]]), method="nearest"
        )
        figures["bottom_marginal"] = figure(
            plot_width=self.app_main_size,
            plot_height=200,
            title=None,
            x_range=figures["main"].x_range,
            y_range=(np.min(bottom_marginal.values), np.max(bottom_marginal.values)),
            x_axis_location="above",
            toolbar_location=None,
            tools=[],
        )
        plots["bottom_marginal"] = figures["bottom_marginal"].line(
            x=bottom_marginal.coords[arr.dims[0]].values, y=bottom_marginal.values
        )
        plots["bottom_marginal_err"] = figures["bottom_marginal"].patch(
            x=[], y=[], color=error_fill, fill_alpha=error_alpha, line_color=None
        )

        # Create the right marginal plot
        right_image = arr.sel(**dict([[arr.dims[0], self.cursor[0]]]), method="nearest")
        prepped_right_image = self.prep_image(right_image)
        self.app_context["color_maps"]["right"] = LinearColorMapper(
            default_palette,
            low=np.min(prepped_right_image),
            high=np.max(prepped_right_image),
            nan_color="black",
        )
        figures["right"] = figure(
            plot_width=self.app_marginal_size,
            plot_height=self.app_main_size,
            title=None,
            x_range=figures["z_marginal"].x_range,
            y_range=figures["main"].y_range,
            toolbar_location=None,
            tools=[],
        )
        figures["right"].yaxis.major_label_text_font_size = "0pt"
        plots["right"] = figures["right"].image(
            [prepped_right_image],
            x=self.app_context["data_range"]["z"][0],
            y=self.app_context["data_range"]["y"][0],
            dw=self.app_context["data_range"]["z"][1] - self.app_context["data_range"]["z"][0],
            dh=self.app_context["data_range"]["y"][1] - self.app_context["data_range"]["y"][0],
            color_mapper=self.app_context["color_maps"]["right"],
        )
        right_marginal = right_image.sel(**dict([[arr.dims[2], self.cursor[2]]]), method="nearest")
        figures["right_marginal"] = figure(
            plot_width=200,
            plot_height=self.app_main_size,
            title=None,
            y_range=figures["main"].y_range,
            x_range=(np.min(right_marginal.values), np.max(right_marginal.values)),
            y_axis_location="left",
            toolbar_location=None,
            tools=[],
        )
        plots["right_marginal"] = figures["right_marginal"].line(
            y=right_marginal.coords[arr.dims[1]].values, x=right_marginal.values
        )
        plots["right_marginal_err"] = figures["right_marginal"].patch(
            x=[], y=[], color=error_fill, fill_alpha=error_alpha, line_color=None
        )

        cursor_lines = self.add_cursor_lines(figures["main"])

        # Attach tools and callbacks
        toggle = widgets.Toggle(label="Show Stat. Variation", button_type="success", active=False)

        def set_show_stat_variation(should_show):
            self.app_context["show_stat_variation"] = should_show

            if should_show:
                main_image_data = arr.sel(**dict([[arr.dims[2], self.cursor[2]]]), method="nearest")
                update_stat_variation(
                    "z",
                    arr.sel(
                        **dict([[arr.dims[0], self.cursor[0]], [arr.dims[1], self.cursor[1]]]),
                        method="nearest"
                    ),
                )
                update_stat_variation(
                    "bottom",
                    main_image_data.sel(**dict([[arr.dims[1], self.cursor[1]]]), method="nearest"),
                )
                update_stat_variation(
                    "right",
                    main_image_data.sel(**dict([[arr.dims[0], self.cursor[0]]]), method="nearest"),
                )
                plots["z_marginal_err"].visible = True
                plots["bottom_marginal_err"].visible = True
                plots["right_marginal_err"].visible = True
            else:
                plots["z_marginal_err"].visible = False
                plots["bottom_marginal_err"].visible = False
                plots["right_marginal_err"].visible = False

        toggle.on_click(set_show_stat_variation)

        scan_keys = ["x", "y", "z", "pass_energy", "hv", "location", "id", "probe_pol", "pump_pol"]
        scan_info_source = ColumnDataSource(
            {
                "keys": [k for k in scan_keys if k in arr.attrs],
                "values": [
                    str(v) if isinstance(v, float) and np.isnan(v) else v
                    for v in [arr.attrs[k] for k in scan_keys if k in arr.attrs]
                ],
            }
        )
        scan_info_columns = [
            widgets.TableColumn(field="keys", title="Attr."),
            widgets.TableColumn(field="values", title="Value"),
        ]

        POINTER_MODES = [
            ("Cursor", "cursor"),
            ("Path", "path"),
        ]

        COLOR_MODES = [
            ("Adaptive Hist. Eq. (Slow)", "adaptive_equalization"),
            # ('Histogram Eq.', 'equalization',), # not implemented
            ("Linear", "linear"),
            # ('Log', 'log',), # not implemented
        ]

        def on_change_color_mode(event):
            new_color_mode = event.item
            self.app_context["color_mode"] = new_color_mode
            if old is None or old != new_color_mode:
                cursor = self.cursor
                right_image_data = arr.sel(**dict([[arr.dims[0], cursor[0]]]), method="nearest")
                bottom_image_data = arr.sel(**dict([[arr.dims[1], cursor[1]]]), method="nearest")
                main_image_data = arr.sel(**dict([[arr.dims[2], cursor[2]]]), method="nearest")
                prepped_right_image = self.prep_image(right_image_data)
                prepped_bottom_image = self.prep_image(bottom_image_data)
                prepped_main_image = self.prep_image(main_image_data)
                plots["right"].data_source.data = {"image": [prepped_right_image]}
                plots["bottom"].data_source.data = {"image": [prepped_bottom_image.T]}
                plots["main"].data_source.data = {"image": [prepped_main_image.T]}
                update_right_colormap(None, None, right_color_range_slider.value)
                update_bottom_colormap(None, None, bottom_color_range_slider.value)
                update_main_colormap(None, None, main_color_range_slider.value)

        color_mode_dropdown = widgets.Dropdown(
            label="Color Mode", button_type="primary", menu=COLOR_MODES
        )
        color_mode_dropdown.on_click(on_change_color_mode)

        symmetry_point_name_input = widgets.TextInput(title="Symmetry Point Name", value="G")
        snap_checkbox = widgets.CheckboxButtonGroup(labels=["Snap Axes"], active=[])
        place_symmetry_point_at_cursor_button = widgets.Button(
            label="Place Point", button_type="primary"
        )

        def update_symmetry_points_for_display():
            pass

        def place_symmetry_point():
            cursor_dict = dict(zip(arr.dims, self.cursor))
            skip_dimensions = {"eV", "delay", "cycle"}
            if "symmetry_points" not in arr.attrs:
                arr.attrs["symmetry_points"] = {}

            snap_distance = {
                "phi": 2,
                "beta": 2,
                "kx": 0.01,
                "ky": 0.01,
                "kz": 0.01,
                "kp": 0.01,
                "hv": 4,
            }

            cursor_dict = {k: v for k, v in cursor_dict.items() if k not in skip_dimensions}
            snapped = copy.copy(cursor_dict)

            if "Snap Axes" in [snap_checkbox.labels[i] for i in snap_checkbox.active]:
                for axis, value in cursor_dict.items():
                    options = [
                        point[axis]
                        for point in arr.attrs["symmetry_points"].values()
                        if axis in point
                    ]
                    options = sorted(options, key=lambda x: np.abs(x - value))
                    if options and np.abs(options[0] - value) < snap_distance[axis]:
                        snapped[axis] = options[0]

            arr.attrs["symmetry_points"][symmetry_point_name_input.value] = snapped

        place_symmetry_point_at_cursor_button.on_click(place_symmetry_point)

        main_color_range_slider = widgets.RangeSlider(
            start=0,
            end=100,
            value=(
                0,
                100,
            ),
            title="Color Range (Main)",
        )
        right_color_range_slider = widgets.RangeSlider(
            start=0,
            end=100,
            value=(
                0,
                100,
            ),
            title="Color Range (%s Marginal)" % arr.dims[1],
        )
        bottom_color_range_slider = widgets.RangeSlider(
            start=0,
            end=100,
            value=(
                0,
                100,
            ),
            title="Color Range (%s Marginal)" % arr.dims[0],
        )
        layout = row(
            column(figures["main"], figures["bottom"], figures["bottom_marginal"]),
            column(
                figures["right"],
                figures["z_marginal"],
                Spacer(width=self.app_marginal_size, height=200),
            ),
            column(
                figures["right_marginal"],
                Spacer(width=200, height=self.app_marginal_size),
                Spacer(width=200, height=200),
            ),
            column(
                column(
                    widgets.Dropdown(
                        label="Pointer Mode", button_type="primary", menu=POINTER_MODES
                    )
                ),
                widgets.Tabs(
                    tabs=[
                        widgets.Panel(
                            child=column(
                                Div(text="<h2>Colorscale:</h2>"),
                                color_mode_dropdown,
                                main_color_range_slider,
                                right_color_range_slider,
                                bottom_color_range_slider,
                                Div(text='<h2 style="padding-top: 30px;">General Settings:</h2>'),
                                toggle,
                                self._cursor_info,
                                sizing_mode="scale_width",
                            ),
                            title="Settings",
                        ),
                        widgets.Panel(
                            child=column(
                                app_widgets["info_div"],
                                Div(
                                    text='<h2 style="padding-top: 30px; padding-bottom: 10px;">Scan Info</h2>'
                                ),
                                widgets.DataTable(
                                    source=scan_info_source,
                                    columns=scan_info_columns,
                                    width=400,
                                    height=400,
                                ),
                                sizing_mode="scale_width",
                                width=400,
                            ),
                            title="Info",
                        ),
                        widgets.Panel(
                            child=column(
                                Div(text="<h2>Preparation</h2>"),
                                symmetry_point_name_input,
                                snap_checkbox,
                                place_symmetry_point_at_cursor_button,
                                sizing_mode="scale_width",
                            ),
                            title="Preparation",
                        ),
                    ],
                    width=400,
                ),
            ),
        )

        update_main_colormap = self.update_colormap_for("main")
        update_bottom_colormap = self.update_colormap_for("bottom")
        update_right_colormap = self.update_colormap_for("right")

        def click_z_marginal(event):
            self.cursor = [self.cursor[0], self.cursor[1], event.x]
            cursor = self.cursor

            main_image = arr.sel(**dict([[arr.dims[2], cursor[2]]]), method="nearest")
            plots["main"].data_source.data = {"image": [self.prep_image(main_image).T]}
            update_main_colormap(None, None, main_color_range_slider.value)
            right_marginal_data = main_image.sel(
                **dict([[arr.dims[0], cursor[0]]]), method="nearest"
            )
            bottom_marginal_data = main_image.sel(
                **dict([[arr.dims[1], cursor[1]]]), method="nearest"
            )
            plots["bottom_marginal"].data_source.data = {
                "x": bottom_marginal_data.coords[arr.dims[0]].values,
                "y": bottom_marginal_data.values,
            }
            plots["right_marginal"].data_source.data = {
                "y": right_marginal_data.coords[arr.dims[1]].values,
                "x": right_marginal_data.values,
            }
            if self.app_context["show_stat_variation"]:
                update_stat_variation("right", right_marginal_data)
                update_stat_variation("bottom", bottom_marginal_data)
            figures["bottom_marginal"].y_range.start = np.min(bottom_marginal_data.values)
            figures["bottom_marginal"].y_range.end = np.max(bottom_marginal_data.values)
            figures["right_marginal"].x_range.start = np.min(right_marginal_data.values)
            figures["right_marginal"].x_range.end = np.max(right_marginal_data.values)

        def click_main_image(event):
            self.cursor = [event.x, event.y, self.cursor[2]]
            cursor = self.cursor
            right_image_data = arr.sel(**dict([[arr.dims[0], cursor[0]]]), method="nearest")
            bottom_image_data = arr.sel(**dict([[arr.dims[1], cursor[1]]]), method="nearest")
            prepped_right_image = self.prep_image(right_image_data)
            prepped_bottom_image = self.prep_image(bottom_image_data)
            plots["right"].data_source.data = {"image": [prepped_right_image]}
            plots["bottom"].data_source.data = {"image": [prepped_bottom_image.T]}
            update_right_colormap(None, None, right_color_range_slider.value)
            update_bottom_colormap(None, None, bottom_color_range_slider.value)
            right_marginal_data = right_image_data.sel(
                **dict([[arr.dims[2], cursor[2]]]), method="nearest"
            )
            bottom_marginal_data = bottom_image_data.sel(
                **dict([[arr.dims[2], cursor[2]]]), method="nearest"
            )
            z_data = arr.sel(
                **dict([[arr.dims[0], cursor[0]], [arr.dims[1], cursor[1]]]), method="nearest"
            )
            plots["z_marginal"].data_source.data = {
                "x": z_coords.values,
                "y": z_data.values,
            }
            plots["bottom_marginal"].data_source.data = {
                "x": bottom_marginal_data.coords[arr.dims[0]].values,
                "y": bottom_marginal_data.values,
            }
            plots["right_marginal"].data_source.data = {
                "y": right_marginal_data.coords[arr.dims[1]].values,
                "x": right_marginal_data.values,
            }
            if self.app_context["show_stat_variation"]:
                update_stat_variation("z", z_data)
                update_stat_variation("right", right_marginal_data)
                update_stat_variation("bottom", bottom_marginal_data)
            figures["z_marginal"].y_range.start = np.min(z_data.values)
            figures["z_marginal"].y_range.end = np.max(z_data.values)
            figures["bottom_marginal"].y_range.start = np.min(bottom_marginal_data.values)
            figures["bottom_marginal"].y_range.end = np.max(bottom_marginal_data.values)
            figures["right_marginal"].x_range.start = np.min(right_marginal_data.values)
            figures["right_marginal"].x_range.end = np.max(right_marginal_data.values)
            if "z_fit" in plots and arr.dims[2] == "eV":
                z_fit_data = z_data.sel(eV=slice(-0.25, 0.2))
                new_z_fit = GStepBModel().guess_fit(z_fit_data)
                plots["z_fit"].data_source.data = {
                    "x": z_fit_data.coords["eV"].values,
                    "y": new_z_fit.best_fit,
                }
                app_widgets["info_div"].text = info_formatter.format(
                    new_z_fit.params["center"].value * 1000,
                    new_z_fit.params["width"].value * 1000,
                )
            if "z_fit" in plots and arr.dims[2] == "delay":
                try:
                    t0 = float(arr.attrs.get("t0", 0))
                    after_t0 = z_data.sel(delay=slice(t0 - 0.2, None))
                    exp_model = ExponentialDecayCModel()
                    z_fit = exp_model.guess_fit(after_t0, params={"t0": {"value": t0}})
                    plots["z_fit"].data_source.data = {
                        "x": after_t0.coords["delay"],
                        "y": z_fit.best_fit,
                    }
                except Exception as e:
                    plots["z_fit"].data_source.data = {
                        "x": [],
                        "y": [],
                    }

            self.save_app()

        figures["z_marginal"].on_event(events.Tap, click_z_marginal)
        figures["main"].on_event(events.Tap, click_main_image)
        main_color_range_slider.on_change("value", update_main_colormap)
        bottom_color_range_slider.on_change("value", update_bottom_colormap)
        right_color_range_slider.on_change("value", update_right_colormap)

        doc.add_root(layout)
        doc.title = "Bokeh Tool"
        self.load_app()
        self.save_app()



"""Provides the core IO facilities supported by PyARPES.

The most important here are the data loading functions (load_data, load_example_data).
and pickling utilities.

Heavy lifting is actually performed by the plugin definitions which know how to ingest
different data formats into the PyARPES data model.
"""

import warnings
from dataclasses import dataclass
from pathlib import Path

import numpy as np
import pandas as pd
import xarray as xr

from arpes.endstations import load_scan
from arpes.typing import DataType

__all__ = (
    "load_data",
    "load_folder",
    "export_dataset",
    "load_example_data",
    "list_pickles",
    "stitch",
)


def load_data(file: str | Path | int, location: str | type, **kwargs) -> xr.Dataset:
    """Load a piece of data using available plugins.

    Args
    ----
    file : An identifier for the file which should be loaded. If this is a number or
        can be coerced to one, data will be loaded from the workspace data folder if
        a matching unique file can be found for the number. If the value is a
        relative path, locations relative to the cwd and the workspace data folder
        will be checked. Absolute paths can also be used in a pinch.
    location : The name of the endstation/plugin to use. You should try to provide
        one. If None is provided, the loader will try to find an appropriate one
        based on the file extension and brute force. This will be slower and can be
        error prone in certain circumstances.

        Optionally, you can pass a loading plugin (the class) through this kwarg and
        directly specify the class to be used.


    Returns
    -------
        The loaded data. Ideally, data which is loaded through the plugin system should
        be highly compliant with the PyARPES data model and should work seamlessly with
        PyARPES analysis code.
    """
    try:
        file = int(str(file))
    except ValueError:
        file = str(Path(file).absolute())

    desc = {
        "file": file,
        "location": location,
    }

    return load_scan(desc, **kwargs)


def load_folder(
    folder: Path, location: str | type, pattern: str = "*", **kwargs
) -> list[xr.Dataset]:
    """Load all files in a folder.

    Args
    ----
    folder : Path
        The folder to load all data from
    location : str | type
        The endstation plugin to use for loading the data
    pattern : str, (default '*')
        The pattern to use for globbing the files in the folder. Defaults to all files

    Returns
    -------
        A list of all the datasets loaded from the specified folder
    """

    all_datasets = []
    for file in folder.glob(pattern):
        all_datasets.append(load_data(file, location, **kwargs))
    return all_datasets


def load_laue(path: Path | str):
    """Load NorthStart Laue backscattering data."""
    northstar_62_69_dtype = np.dtype(
        [
            ("pad1", "B", (2364,)),  # unused
            ("sample", "S52"),
            ("user", "S52"),
            ("comment", "S512"),
            ("pad2", "B", (228,)),  # unused
        ]
    )
    """
    Primitive support for loading Laue data from the NorthStar x-ray backscattering DAQ program.

    Laue file structure courtesy Jonathan Denlinger, MERLIN endstations at the ALS
    16-bit binary Laue histogram (.hs2) file
    Format:  2 byte*256*256= 131072 long + header info at the end
    Northstar 6.0:  132028 bytes   (956 byte extra)
    Northstar 6.2.6.9:  134280 bytes  ( 3208 byte extra)
    header includes (offset from start of header):
    byte 65536 / 131072 / 2364   = sample name (character string - read to double space)
    byte 65587 / 131124 / 2416  = operator (character string - read to double space)
    byte 65638 / 131176 / 2468  = date in mm/dd/yy format (8 character string)
    byte 65811 / 131760 / 2984  = dwell time * 10 in seconds (word)
    byte 65821 / 131776 / 3000  = mA
    byte 65823 / 131780 / 3004  = kV
    byte 131664 / 592 = index file name
    """
    if isinstance(path, str):
        path = Path(path)

    binary_data = path.read_bytes()
    table, header = binary_data[:131072], binary_data[131072:]

    table = np.fromstring(table, dtype=np.uint16).reshape(256, 256)
    header = np.fromstring(header, dtype=northstar_62_69_dtype).item()

    arr = xr.DataArray(
        table,
        coords={"x": np.array(range(256)), "y": np.array(range(256))},
        dims=[
            "x",
            "y",
        ],
        attrs={
            "sample": header[1].split(b"\0")[0].decode("ascii"),
            "user": header[2].split(b"\0")[0].decode("ascii"),
            "comment": header[3].split(b"\0")[0].decode("ascii"),
        },
    )
    return arr


DATA_EXAMPLES = {
    "cut": ("ALG-MC", "cut.fits"),
    "map": ("example_data", "fermi_surface.nc"),
    "photon_energy": ("example_data", "photon_energy.nc"),
    "nano_xps": ("example_data", "nano_xps.nc"),
    "temperature_dependence": ("example_data", "temperature_dependence.nc"),
}


def load_example_data(example_name="cut") -> xr.Dataset:
    """Provide sample data for executable documentation."""
    if example_name not in DATA_EXAMPLES:
        warnings.warn(
            f"Could not find requested example_name: {example_name}. Please provide one of {list(DATA_EXAMPLES.keys())}"
        )

    location, example = DATA_EXAMPLES[example_name]
    file = Path(__file__).parent / "example_data" / example
    return load_data(file=file, location=location)


@dataclass
class ExampleData:
    @property
    def cut(self) -> xr.Dataset:
        return load_example_data("cut")

    @property
    def map(self) -> xr.Dataset:
        return load_example_data("map")

    @property
    def photon_energy(self) -> xr.Dataset:
        return load_example_data("photon_energy")

    @property
    def nano_xps(self) -> xr.Dataset:
        return load_example_data("nano_xps")

    @property
    def temperature_dependence(self) -> xr.Dataset:
        return load_example_data("temperature_dependence")


example_data = ExampleData()


def stitch(
    df_or_list: list[str] | pd.DataFrame,
    attr_or_axis: str,
    built_axis_name: str = None,
    sort: bool = True,
) -> DataType:
    """Stitches together a sequence of scans or a DataFrame.

    Args:
        df_or_list: The list of the files to load
        attr_or_axis: Coordinate or attribute in order to promote to an index. I.e. if 't_a' is specified,
                      we will create a new axis corresponding to the temperature and concatenate the data along this axis
        built_axis_name: The name of the concatenated output dimensions
        sort: Whether to sort inputs to the concatenation according to their `attr_or_axis` value.

    Returns:
        The concatenated data.
    """
    list_of_files = None
    if isinstance(df_or_list, (pd.DataFrame,)):
        list_of_files = list(df_or_list.index)
    else:
        if not isinstance(df_or_list, (list, tuple)):
            raise TypeError(
                "Expected an interable for a list of the scans to stitch together"
            )

        list_of_files = list(df_or_list)

    built_axis_name = attr_or_axis if built_axis_name is None else built_axis_name

    if not list_of_files:
        raise ValueError("Must supply at least one file to stitch")

    loaded = [
        f if isinstance(f, (xr.DataArray, xr.Dataset)) else load_data(f)
        for f in list_of_files
    ]

    for i, loaded_file in enumerate(loaded):
        value = None
        if isinstance(attr_or_axis, (list, tuple)):
            value = attr_or_axis[i]
        elif attr_or_axis in loaded_file.attrs:
            value = loaded_file.attrs[attr_or_axis]
        elif attr_or_axis in loaded_file.coords:
            value = loaded_file.coords[attr_or_axis]

        loaded_file = loaded_file.assign_coords(dict([[built_axis_name, value]]))

    if sort:
        loaded.sort(key=lambda x: x.coords[built_axis_name])

    concatenated = xr.concat(loaded, dim=built_axis_name)
    if "id" in concatenated.attrs:
        del concatenated.attrs["id"]

    from arpes.provenance import provenance_multiple_parents

    provenance_multiple_parents(
        concatenated,
        loaded,
        {
            "what": "Stitched together separate datasets",
            "by": "stitch",
            "dim": built_axis_name,
        },
    )

    return concatenated


def export_dataset(dataset: xr.Dataset, path: str | Path) -> None:
    """Correct bad keys/values then export dataset to netcdf.

    Note that all values that are not strings, ints, or floats are converted to strings.

    Args
    ----
        dataset : xr.Dataset
            The dataset to export
        path : str | Path
            export path
    """
    path = Path(path) if isinstance(path, str) else path
    dataset = dataset.copy(deep=True)

    fixed_values = {}
    bad_keys = []
    for key, value in dataset.attrs.items():
        if not isinstance(value, (str, int, float)) or isinstance(value, bool):
            fixed_values[key] = str(value)
        if "/" in key:
            bad_keys.append(key)
    for key in fixed_values:
        dataset.attrs[key] = fixed_values[key]
    for key in bad_keys:
        dataset.attrs[key.replace("/", " per ")] = dataset.attrs.pop(key)

    if path.suffix != ".nc":
        warnings.warn(
            "The provided path doesn't have a .nc extension. Adding one and continuing."
        )
        path = path.parent / f"{path.name}.nc"
    if not path.parent.exists():
        path.parent.mkdir()
    dataset.to_netcdf(path)


# TODO: J: add an import dataset function which converts bools/Nones back to their original types



"""Contains routines for calculating and removing the classic Shirley background."""
import warnings

import numpy as np
import xarray as xr

from arpes.provenance import update_provenance
from arpes.typing import DataType
from arpes.utilities import normalize_to_spectrum

__all__ = (
    "calculate_shirley_background",
    "calculate_shirley_background_full_range",
    "remove_shirley_background",
)


@update_provenance("Remove Shirley background")
def remove_shirley_background(xps: DataType, **kwargs) -> DataType:
    """Calculates and removes a Shirley background from a spectrum.

    Only the background corrected spectrum is retrieved.

    Args:
        xps: The input array.
        kwargs: Parameters to feed to the background estimation routine.

    Returns:
        The the input array with a Shirley background subtracted.
    """
    xps = normalize_to_spectrum(xps)
    return xps - calculate_shirley_background(xps, **kwargs)


def _calculate_shirley_background_full_range(
    xps: np.ndarray, eps=1e-7, max_iters=50, n_samples=5
) -> np.ndarray:
    """Core routine for calculating a Shirley background on np.ndarray data."""
    background = np.copy(xps)
    cumulative_xps = np.cumsum(xps, axis=0)
    total_xps = np.sum(xps, axis=0)

    rel_error = np.inf

    i_left = np.mean(xps[:n_samples], axis=0)
    i_right = np.mean(xps[-n_samples:], axis=0)

    iter_count = 0

    k = i_left - i_right
    for iter_count in range(max_iters):
        cumulative_background = np.cumsum(background, axis=0)
        total_background = np.sum(background, axis=0)

        new_bkg = np.copy(background)

        for i in range(len(new_bkg)):
            new_bkg[i] = i_right + k * (
                (total_xps - cumulative_xps[i] - (total_background - cumulative_background[i]))
                / (total_xps - total_background + 1e-5)
            )

        rel_error = np.abs(np.sum(new_bkg, axis=0) - total_background) / (total_background)

        background = new_bkg

        if np.any(rel_error < eps):
            break

    if (iter_count + 1) == max_iters:
        warnings.warn(
            "Shirley background calculation did not converge "
            + "after {} steps with relative error {}!".format(max_iters, rel_error)
        )

    return background


@update_provenance("Calculate full range Shirley background")
def calculate_shirley_background_full_range(
    xps: DataType, eps=1e-7, max_iters=50, n_samples=5
) -> DataType:
    """Calculates a shirley background.

    The background is defined according to:

    S(E) = I(E_right) + k * (A_right(E)) / (A_left(E) + A_right(E))

    Typically

    k := I(E_right) - I(E_left)

    The iterative method is continued so long as the total background is not converged to relative error
    `eps`.

    The method continues for a maximum number of iterations `max_iters`.

    In practice, what we can do is to calculate the cumulative sum of the data along the energy axis of
    both the data and the current estimate of the background

    Args:
        xps: The input data.
        eps: Convergence parameter.
        max_iters: The maximum number of iterations to allow before convengence.
        n_samples: The number of samples to use at the boundaries of the input data.

    Returns:
        A monotonic Shirley backgruond over the entire energy range.
    """
    xps = normalize_to_spectrum(xps).copy(deep=True)
    core_dims = [d for d in xps.dims if d != "eV"]

    return xr.apply_ufunc(
        _calculate_shirley_background_full_range,
        xps,
        eps,
        max_iters,
        n_samples,
        input_core_dims=[core_dims, [], [], []],
        output_core_dims=[core_dims],
        exclude_dims=set(core_dims),
        vectorize=False,
    )


@update_provenance("Calculate limited range Shirley background")
def calculate_shirley_background(
    xps: DataType, energy_range: slice = None, eps=1e-7, max_iters=50, n_samples=5
) -> DataType:
    """Calculates a shirley background iteratively over the full energy range `energy_range`.

    Uses `calculate_shirley_background_full_range` internally.

    Outside the indicated range, the background is extrapolated as a constant from
    the nearest in-range value.

    Args:
        xps: The input data.
        energy_range: A slice with the energy range to be used.
        eps: Convergence parameter.
        max_iters: The maximum number of iterations to allow before convengence.
        n_samples: The number of samples to use at the boundaries of the input data.

    Returns:
        A monotonic Shirley backgruond over the entire energy range.
    """
    if energy_range is None:
        energy_range = slice(None, None)

    xps = normalize_to_spectrum(xps)
    xps_for_calc = xps.sel(eV=energy_range)

    bkg = calculate_shirley_background_full_range(xps_for_calc, eps, max_iters, n_samples)
    bkg = bkg.transpose(*xps.dims)
    full_bkg = xps * 0

    left_idx = np.searchsorted(full_bkg.eV.values, bkg.eV.values[0], side="left")
    right_idx = left_idx + len(bkg)

    full_bkg.values[:left_idx] = bkg.values[0]
    full_bkg.values[left_idx:right_idx] = bkg.values
    full_bkg.values[right_idx:] = bkg.values[-1]

    return full_bkg



"""Utilities for modifying, iterating over, and transforming dictionaries."""
import re
from collections import OrderedDict

from arpes.utilities.xarray import lift_dataarray_attrs, lift_datavar_attrs
from typing import Any, Dict, Union

__all__ = (
    "rename_keys",
    "clean_keys",
    "rename_dataarray_attrs",
    "clean_datavar_attribute_names",
    "clean_attribute_names",
    "case_insensitive_get",
)


def _rename_key(d: Union[Dict[str, Any], OrderedDict], k: str, nk: str) -> None:
    if k in d:
        d[nk] = d[k]
        del d[k]


def rename_keys(
    d: Union[Dict[str, Any], OrderedDict], keys_dict: Dict[str, str]
) -> Union[Dict[str, Any], OrderedDict]:
    """Renames all the keys of `d` according to the remapping in `keys_dict`."""
    d = d.copy()
    for k, nk in keys_dict.items():
        _rename_key(d, k, nk)

    return d


def clean_keys(d):
    """Renames dictionary keys so that they are more Pythonic."""

    def clean_single_key(k):
        k = k.replace(" ", "_")
        k = k.replace(".", "_")
        k = k.lower()
        k = re.sub(r"[()/?]", "", k)
        k = k.replace("__", "_")
        return k

    return dict(zip([clean_single_key(k) for k in d.keys()], d.values()))


def case_insensitive_get(d: dict, key: str, default=None, take_first=False):
    """Looks up a key in a dictionary ignoring case.

    We use this sometimes to be nicer to users who don't provide perfectly sanitized data.

    Args:
        d: The dictionary to perform lookup in
        key: The key to get
        default: A default value if the key is not present
        take_first: Whether to take the first entry if there were multiple found
    """
    found_value = False
    value = None

    for k, v in d.items():
        if k.lower() == key.lower():
            if not take_first and found_value:
                raise ValueError("Duplicate case insensitive keys")

            value = v
            found_value = True

            if take_first:
                break

    if not found_value:
        return default

    return value


rename_dataarray_attrs = lift_dataarray_attrs(rename_keys)
clean_attribute_names = lift_dataarray_attrs(clean_keys)

clean_datavar_attribute_names = lift_datavar_attrs(clean_keys)



"""Defines models useful for studying excited carriers in Tr-ARPES."""
from lmfit.models import update_param_vals
import numpy as np
import lmfit as lf

from .x_model_mixin import XModelMixin

__all__ = ["ExponentialDecayCModel", "TwoExponentialDecayCModel"]


class ExponentialDecayCModel(XModelMixin):
    """A model for fitting an exponential decay with a constant background."""

    @staticmethod
    def exponential_decay_c(x, amp, tau, t0, const_bkg):
        """Represents an exponential decay after a point (delta) impulse.

        This coarsely models the dynamics after excitation in a
        pump-probe experiment.

        Args:
            x
            amp
            tau
            t0
            const_bkg

        Returns:
            The decay profile.
        """
        dx = x - t0
        mask = (dx >= 0) * 1
        return const_bkg + amp * mask * np.exp(-(x - t0) / tau)

    def __init__(self, independent_vars=("x",), prefix="", missing="raise", name=None, **kwargs):
        """Defer to lmfit for initialization."""
        kwargs.update({"prefix": prefix, "missing": missing, "independent_vars": independent_vars})
        super().__init__(self.exponential_decay_c, **kwargs)

        # amp is also a parameter, but we have no hint for it
        self.set_param_hint("tau", min=0.0)
        # t0 is also a parameter, but we have no hint for it
        self.set_param_hint("const_bkg")

    def guess(self, data, x=None, **kwargs):
        """Make heuristic estimates of parameters.

        200fs is a reasonable value for the time constant, in fact its probably a bit large.
        We assume data is probably calibrated so that t0 is at 0 delay.
        """
        pars = self.make_params()

        pars["%stau" % self.prefix].set(value=0.2)  # 200fs
        pars["%st0" % self.prefix].set(value=0)
        pars["%sconst_bkg" % self.prefix].set(value=data.mean())
        pars["%samp" % self.prefix].set(value=data.max() - data.mean())

        return update_param_vals(pars, self.prefix, **kwargs)

    __init__.doc = lf.models.COMMON_INIT_DOC
    guess.__doc__ = lf.models.COMMON_GUESS_DOC


class TwoExponentialDecayCModel(XModelMixin):
    """A model for fitting an exponential decay with a constant background."""

    @staticmethod
    def twoexponential_decay_c(x, amp, t0, tau1, tau2, const_bkg):
        """Like `exponential_decay_c`, except with two timescales.

        This is meant to model if two different quasiparticle decay channels are allowed,
        represented by `tau1` and `tau2`.
        """
        dx = x - t0
        mask = (dx >= 0) * 1
        y = const_bkg + amp * (1 - np.exp(-dx / tau1)) * np.exp(-dx / tau2)
        f = y.copy()
        f[dx < 0] = const_bkg
        f[dx >= 0] = y[dx >= 0]
        return f

    def __init__(self, independent_vars=("x",), prefix="", missing="raise", name=None, **kwargs):
        """Defer to lmfit for initialization."""
        kwargs.update({"prefix": prefix, "missing": missing, "independent_vars": independent_vars})
        super().__init__(self.twoexponential_decay_c, **kwargs)

        # amp is also a parameter, but we have no hint for it
        self.set_param_hint("tau1", min=0.0)
        self.set_param_hint("tau2", min=0.0)
        # t0 is also a parameter, but we have no hint for it
        self.set_param_hint("const_bkg")

    def guess(self, data, x=None, **kwargs):
        """Placeholder for making better heuristic guesses here."""
        pars = self.make_params()

        pars["%stau1" % self.prefix].set(value=0.2)  # 200fs
        pars["%stau2" % self.prefix].set(value=1)  # 1ps
        pars["%st0" % self.prefix].set(value=0)
        pars["%sconst_bkg" % self.prefix].set(value=data.mean())
        pars["%samp" % self.prefix].set(value=data.max() - data.mean())

        return update_param_vals(pars, self.prefix, **kwargs)

    __init__.doc = lf.models.COMMON_INIT_DOC
    guess.__doc__ = lf.models.COMMON_GUESS_DOC



"""Provides standard corrections for datasets.

Largely, this covers:
1. Fermi edge corrections
2. Background estimation and subtraction

It also contains utilities related to identifying a piece of data
earlier in a dataset which can be used to furnish equivalent references.

"""
from collections import OrderedDict

from arpes.typing import DataType
from arpes.utilities import normalize_to_dataset, deep_equals
from .fermi_edge_corrections import *


__all__ = (
    "reference_key",
    "correction_from_reference_set",
)


class HashableDict(OrderedDict):
    """Implements hashing for ordered dictionaries.

    The dictionary must be ordered for the hash to be stable.
    """

    def __hash__(self):
        return hash(frozenset(self.items()))


def reference_key(data: DataType):
    """Calculates a key/hash for data determining reference/correction equality."""
    data = normalize_to_dataset(data)

    return HashableDict(data.S.reference_settings)


def correction_from_reference_set(data: DataType, reference_set):
    """Determines which correction to use from a set of references."""
    data = normalize_to_dataset(data)

    correction = None
    for k, corr in reference_set.items():
        if deep_equals(dict(reference_key(data)), dict(k)):
            correction = corr
            break

    return correction



"""This package contains utilities related to taking more complicated shaped selections around data.

Currently it houses just utilities for forming disk and annular selections out of data.
"""

from typing import Dict, Optional, Tuple, Union

import numpy as np
import xarray as xr

from arpes.typing import DataType
from arpes.utilities import normalize_to_spectrum
from arpes.utilities.xarray import unwrap_xarray_dict

__all__ = ("select_disk", "select_disk_mask", "unravel_from_mask", "ravel_from_mask")


def ravel_from_mask(data, mask):
    """Selects out the data from a ND array whose points are marked true in `mask`.

    See also `unravel_from_mask`
    below which allows you to write back into data after you have transformed the 1D output in some way.

    These two functions are especially useful for hierarchical curve fitting where you want to rerun a fit over a subset
    of the data with a different model, such as when you know some of the data is best described by two bands rather than
    one.

    Args:
        data:
        mask:

    Returns:
        Raveled data with masked points removed.
    """
    return data.stack(stacked=mask.dims).where(mask.stack(stacked=mask.dims), drop=True)


def unravel_from_mask(template, mask, values, default=np.nan):
    """Creates an array from a mask and a flat collection of the unmasked values.

    Inverse to `ravel_from_mask`, so look at that function as well.

    Args:
        template
        mask
        values
        default

    Returns:
        Unraveled data with default values filled in where the raveled list is missing from the mask.
    """
    dest = template * 0 + 1
    dest_mask = np.logical_not(
        np.isnan(
            template.stack(stacked=template.dims).where(mask.stack(stacked=template.dims)).values
        )
    )
    dest = (dest * default).stack(stacked=template.dims)
    dest.values[dest_mask] = values
    return dest.unstack("stacked")


def _normalize_point(data, around, **kwargs):
    collected_kwargs = {k: kwargs[k] for k in data.dims if k in kwargs}

    if around:
        if isinstance(around, xr.Dataset):
            around = unwrap_xarray_dict({d: around[d] for d in data.dims})
    else:
        around = collected_kwargs

    assert set(around.keys()) == set(data.dims)
    return around


def select_disk_mask(
    data: DataType,
    radius,
    outer_radius=None,
    around: Optional[Union[Dict, xr.Dataset]] = None,
    flat=False,
    **kwargs
) -> np.ndarray:
    """A complement to `select_disk` which only generates the mask for the selection.

    Selects the data in a disk around the point described by `around` and `kwargs`. A point is a labelled
    collection of coordinates that matches all of the dimensions of `data`. The coordinates can either be
    passed through a dict as `around`, as the coordinates of a Dataset through `around` or explicitly in
    keyword argument syntax through `kwargs`. The radius for the disk is specified through the required
    `radius` parameter.

    Returns the ND mask that represents the filtered coordinates.

    Args:
        data: The data which should be masked
        radius: The radius of the circle to mask
        outer_radius: The outer radius of an annulus to mask
        around: The location of the center point otherwise specified by `kwargs`
        flat: Whether to return the mask as a 1D (raveled) mask
          (flat=True) or as a ND mask with the same dimensionality as
          the input data (flat=False).
        kwargs: The location of the center point otherwise specified by `around`

    Returns:
        A mask with the same shape as `data`.
    """
    if outer_radius is not None and radius > outer_radius:
        radius, outer_radius = outer_radius, radius

    data = normalize_to_spectrum(data)
    around = _normalize_point(data, around, **kwargs)

    raveled = data.G.ravel()

    dim_order = list(around.keys())
    dist = np.sqrt(
        np.sum(np.stack([(raveled[d] - around[d]) ** 2 for d in dim_order], axis=1), axis=1)
    )

    mask = dist <= radius
    if outer_radius is not None:
        mask = np.logical_or(mask, dist > outer_radius)

    if flat:
        return mask

    return mask.reshape(data.shape[::-1])


def select_disk(
    data: DataType,
    radius,
    outer_radius=None,
    around: Optional[Union[Dict, xr.Dataset]] = None,
    invert=False,
    **kwargs
) -> Tuple[Dict[str, np.ndarray], np.ndarray, np.ndarray]:
    """Selects the data in a disk (or annulus if `outer_radius` is provided) around the point requested.

    The point is specified by `around` and `kwargs`. A point is a labeled collection of coordinates that
    matches all of the dimensions of `data`. The coordinates can either be passed through a dict as
    `around`, as the coordinates of a Dataset through `around` or explicitly in keyword argument
    syntax through `kwargs`. The radius for the disk is specified through the required `radius` parameter.

    Data is returned as a tuple with the type Tuple[Dict[str, np.ndarray], np.ndarray, np.ndarray]
    containing a dictionary with the filtered lists of coordinates, an array with the original data
    values at these coordinates, and finally an array of the distances to the requested point.

    Args:
        data: The data to perform the selection from
        radius: The inner radius of the annulus selection
        outer_radius: The outer radius of the annulus selection
        around: The central point, otherwise specified by `kwargs`
        invert: Whether to invert the mask, i.e. everything but the annulus
        kwargs: The central point, otherwise specified by `around`
    """
    data = normalize_to_spectrum(data)
    around = _normalize_point(data, around, **kwargs)
    mask = select_disk_mask(data, radius, outer_radius=outer_radius, around=around, flat=True)

    if invert:
        mask = np.logical_not(mask)

    # at this point, around is now a dictionary specifying a point to do the selection around
    raveled = data.G.ravel()
    data_arr = raveled["data"]

    dim_order = list(around.keys())
    dist = np.sqrt(
        np.sum(np.stack([(raveled[d] - around[d]) ** 2 for d in dim_order], axis=1), axis=1)
    )

    masked_coords = {d: cs[mask] for d, cs in raveled.items()}
    return masked_coords, masked_coords["data"], dist[mask]



"""Automated utilities for calculating Fermi edge corrections."""

from typing import TYPE_CHECKING
import lmfit as lf
import matplotlib.pyplot as plt
import numpy as np
from scipy.ndimage import gaussian_filter1d

import xarray as xr
from arpes.typing import DataType
from arpes.fits import (
    GStepBModel,
    LinearModel,
    QuadraticModel,
    AffineBroadenedFD,
    broadcast_model,
)
from arpes.provenance import provenance, update_provenance
from arpes.utilities.math import shift_by

if TYPE_CHECKING:
    from lmfit.model import ModelResult


def _exclude_from_set(excluded):
    def exclude(l):
        return list(set(l).difference(excluded))

    return exclude


exclude_hemisphere_axes = _exclude_from_set({"phi", "eV"})
exclude_hv_axes = _exclude_from_set({"hv", "eV"})


__all__ = (
    "build_quadratic_fermi_edge_correction",
    "build_photon_energy_fermi_edge_correction",
    "apply_photon_energy_fermi_edge_correction",
    "apply_quadratic_fermi_edge_correction",
    "build_direct_fermi_edge_correction",
    "apply_direct_fermi_edge_correction",
    "find_e_fermi_linear_dos",
    "fix_fermi_edge",
)


def find_e_fermi_linear_dos(edc, guess=None, plot=False, ax=None):
    """Estimate the Fermi level under the assumption of a linear density of states.

    Does a reasonable job of finding E_Fermi in-situ for graphene/graphite or other materials with a linear DOS near
    the chemical potential. You can provide an initial guess via guess, or one will be chosen half way through the EDC.

    The Fermi level is estimated as the location where the DoS crosses below an estimated background level

    Args:
        edc: Input data
        guess: Approximate location
        plot: Whether to plot the fit, useful for debugging.

    Returns:
        The Fermi edge position.
    """
    if guess is None:
        guess = edc.eV.values[len(edc.eV) // 2]

    edc = edc - np.percentile(edc.values, (20,))[0]
    mask = edc > np.percentile(edc.sel(eV=slice(None, guess)), 20)
    mod = LinearModel().guess_fit(edc[mask])

    chemical_potential = -mod.params["intercept"].value / mod.params["slope"].value

    if plot:
        if ax is None:
            fig, ax = plt.subplots()
        edc.plot(ax=ax)
        ax.axvline(chemical_potential, linestyle="--", color="red")
        ax.axvline(guess, linestyle="--", color="gray")

    return chemical_potential


def apply_direct_fermi_edge_correction(
    arr: xr.DataArray, correction=None, *args, **kwargs
):
    """Applies a direct fermi edge correction stencil."""
    if correction is None:
        correction = build_direct_fermi_edge_correction(arr, *args, **kwargs)

    shift_amount = (
        -correction / arr.G.stride(generic_dim_names=False)["eV"]
    )  # pylint: disable=invalid-unary-operand-type
    energy_axis = list(arr.dims).index("eV")

    correction_axis = list(arr.dims).index(correction.dims[0])

    corrected_arr = xr.DataArray(
        shift_by(
            arr.values, shift_amount, axis=energy_axis, by_axis=correction_axis, order=1
        ),
        arr.coords,
        arr.dims,
        attrs=arr.attrs,
    )

    if "id" in corrected_arr.attrs:
        del corrected_arr.attrs["id"]

    provenance(
        corrected_arr,
        arr,
        {
            "what": "Shifted Fermi edge to align at 0 along hv axis",
            "by": "apply_photon_energy_fermi_edge_correction",
            "correction": list(
                correction.values
                if isinstance(correction, xr.DataArray)
                else correction
            ),
        },
    )

    return corrected_arr


@update_provenance("Build direct Fermi edge correction")
def build_direct_fermi_edge_correction(
    arr: xr.DataArray, fit_limit=0.001, energy_range=None, plot=False, along="phi"
):
    """Builds a direct fermi edge correction stencil.

    This means that fits are performed at each value of the 'phi' coordinate
    to get a list of fits. Bad fits are thrown out to form a stencil.

    This can be used to shift coordinates by the nearest value in the stencil.

    Args:
        arr
        fit_limit
        energy_range
        plot
        along

    Returns:
        The array of fitted edge coordinates.
    """
    if energy_range is None:
        energy_range = slice(-0.1, 0.1)

    exclude_axes = ["eV", along]
    others = [d for d in arr.dims if d not in exclude_axes]
    edge_fit = broadcast_model(
        GStepBModel, arr.sum(others).sel(eV=energy_range), along
    ).results

    def sieve(c, v):
        return v.item().params["center"].stderr < 0.001

    corrections = edge_fit.G.filter_coord(along, sieve).G.map(
        lambda x: x.params["center"].value
    )

    if plot:
        corrections.plot()

    return corrections


def build_quadratic_fermi_edge_correction(
    arr: xr.DataArray, fit_limit=0.001, eV_slice=None, plot=False
) -> lf.model.ModelResult:
    """Calculates a quadratic Fermi edge correction by edge fitting and then quadratic fitting of edges."""
    # TODO improve robustness here by allowing passing in the location of the fermi edge guess
    # We could also do this automatically by using the same method we use for step detection to find the edge of the
    # spectrometer image

    if eV_slice is None:
        approximate_fermi_level = arr.S.find_spectrum_energy_edges().max()
        eV_slice = slice(approximate_fermi_level - 0.4, approximate_fermi_level + 0.4)
    else:
        approximate_fermi_level = 0
    sum_axes = exclude_hemisphere_axes(arr.dims)
    edge_fit = broadcast_model(
        GStepBModel,
        arr.sum(sum_axes).sel(eV=eV_slice),
        "phi",
        params={"center": {"value": approximate_fermi_level}},
    )

    size_phi = len(arr.coords["phi"])
    not_nanny = (np.logical_not(np.isnan(arr)) * 1).sum("eV") > size_phi * 0.30
    condition = np.logical_and(edge_fit.F.s("center") < fit_limit, not_nanny)

    quadratic_corr = QuadraticModel().guess_fit(
        edge_fit.F.p("center"), weights=condition * 1
    )
    if plot:
        edge_fit.F.p("center").plot()
        plt.plot(arr.coords["phi"], quadratic_corr.best_fit)

    return quadratic_corr


@update_provenance("Build photon energy Fermi edge correction")
def build_photon_energy_fermi_edge_correction(
    arr: xr.DataArray, plot=False, energy_window=0.2
):
    """Builds Fermi edge corrections across photon energy (corrects monochromator miscalibration)."""
    edge_fit = broadcast_model(
        GStepBModel,
        arr.sum(exclude_hv_axes(arr.dims)).sel(eV=slice(-energy_window, energy_window)),
        "hv",
    )

    return edge_fit


def apply_photon_energy_fermi_edge_correction(
    arr: xr.DataArray, correction=None, **kwargs
):
    """Applies Fermi edge corrections across photon energy (corrects monochromator miscalibration)."""
    if correction is None:
        correction = build_photon_energy_fermi_edge_correction(arr, **kwargs)

    correction_values = correction.G.map(lambda x: x.params["center"].value)
    if "corrections" not in arr.attrs:
        arr.attrs["corrections"] = {}

    arr.attrs["corrections"]["hv_correction"] = list(correction_values.values)

    shift_amount = -correction_values / arr.G.stride(generic_dim_names=False)["eV"]
    energy_axis = arr.dims.index("eV")
    hv_axis = arr.dims.index("hv")

    corrected_arr = xr.DataArray(
        shift_by(arr.values, shift_amount, axis=energy_axis, by_axis=hv_axis, order=1),
        arr.coords,
        arr.dims,
        attrs=arr.attrs,
    )

    if "id" in corrected_arr.attrs:
        del corrected_arr.attrs["id"]

    provenance(
        corrected_arr,
        arr,
        {
            "what": "Shifted Fermi edge to align at 0 along hv axis",
            "by": "apply_photon_energy_fermi_edge_correction",
            "correction": list(correction_values.values),
        },
    )

    return corrected_arr


def apply_quadratic_fermi_edge_correction(
    arr: xr.DataArray, correction: lf.model.ModelResult = None, offset=None
):
    """Applies a Fermi edge correction using a quadratic fit for the edge."""
    assert isinstance(arr, xr.DataArray)
    if correction is None:
        correction = build_quadratic_fermi_edge_correction(arr)

    if "corrections" not in arr.attrs:
        arr.attrs["corrections"] = {}

    arr.attrs["corrections"]["FE_Corr"] = correction.best_values

    delta_E = arr.coords["eV"].values[1] - arr.coords["eV"].values[0]
    dims = list(arr.dims)
    energy_axis = dims.index("eV")
    phi_axis = dims.index("phi")

    shift_amount_E = correction.eval(x=arr.coords["phi"].values)

    if offset is not None:
        shift_amount_E = shift_amount_E - offset

    shift_amount = -shift_amount_E / delta_E

    corrected_arr = xr.DataArray(
        shift_by(arr.values, shift_amount, axis=energy_axis, by_axis=phi_axis, order=1),
        arr.coords,
        arr.dims,
        attrs=arr.attrs,
    )

    if "id" in corrected_arr.attrs:
        del corrected_arr.attrs["id"]

    provenance(
        corrected_arr,
        arr,
        {
            "what": "Shifted Fermi edge to align at 0",
            "by": "apply_quadratic_fermi_edge_correction",
            "correction": correction.best_values,
        },
    )

    return corrected_arr


def fit_fermi_edge(cut: xr.DataArray) -> xr.DataArray:
    """
    Fits the curved Fermi edge of a cut to a quadratic model.
    """
    edge_region = cut.sel(eV=slice(-0.1, 0.1))

    edge_fits = broadcast_model(AffineBroadenedFD, edge_region, "phi", progress=False)
    quad_fit: ModelResult = QuadraticModel().guess_fit(edge_fits.F.p("fd_center"))
    return xr.DataArray(
        quad_fit.eval(x=edge_region["phi"]),
        dims="phi",
        coords={"phi": edge_region["phi"]},
    )


def fix_fermi_edge(
    dataset: DataType,
    broadcast_fit: bool = True,
    edge_fit: xr.DataArray = None,
    smoothing_sigma: float = 1.0,
) -> DataType:
    """Automatically correct the Fermi edge in a dataset.

    If the measurement was done with a curved slit, the Fermi edge will only be shifted
    to 0. If the measurement was done with a straight slit, the Fermi edge will be
    shifted to 0 and the curved Fermi edge will be corrected (`slit_shape` attr needs to
    be set to `straight`).

    Args
    ----
    dataset : DataType
        The dataset to correct
    broadcast_fit : bool
        Whether to fit the Fermi edge of the summed spectrum and then broadcast to all
        cuts or fit each cut individually.
    edge_fit : DataArray
        A precomputed edge fit to use. Typically generated using `fit_fermi_edge` on
        metal reference data.
    smoothing_sigma : float
        The sigma value for the Gaussian filter used to smooth the integrated edc for
        Fermi edge detection. Increasing this can fix faulty edge detection in noisy
        data.

    Returns
    -------
        A new dataset with the Fermi edge corrected.
    """
    HEMISPHERE_DIMS = {"eV", "phi"}
    original_spectrum: xr.DataArray = (
        dataset.S.spectrum if isinstance(dataset, xr.Dataset) else dataset
    )
    # We want to be able to pass any dataset for convenience, but we only want to
    # correct the Fermi edge for hemisphere data.
    if not all(dim in original_spectrum.dims for dim in HEMISPHERE_DIMS):
        return dataset

    dataset = dataset.copy(deep=True)
    if edge_fit is None:
        integrated_edc = gaussian_filter1d(
            original_spectrum.sum(
                [dim for dim in original_spectrum.dims if dim != "eV"]
            ),
            sigma=smoothing_sigma,
        )
        dropoff_index = np.argmin(np.diff(integrated_edc))
        dataset.coords["eV"] = (
            dataset.coords["eV"] - dataset.coords["eV"].values[dropoff_index]
        )
        if dataset.attrs.get("slit_shape") != "straight":
            return dataset

    spectrum: xr.DataArray = (
        dataset.S.spectrum if isinstance(dataset, xr.Dataset) else dataset
    )

    scan_axes = [dim for dim in spectrum.dims if dim not in HEMISPHERE_DIMS]
    # we need to stack something
    if not scan_axes:
        spectrum = spectrum.expand_dims("temp")
        scan_axes = ["temp"]

    stacked = spectrum.stack(flattened=scan_axes)
    if edge_fit is None:
        edge_fit = (
            fit_fermi_edge(stacked.sum("flattened")) if broadcast_fit is True else None
        )

    def fix_one_fermi_edge(cut: xr.DataArray, edge_fit=None) -> xr.DataArray:
        edge_fit = edge_fit if edge_fit is not None else fit_fermi_edge(cut)
        return cut.G.shift_by(edge_fit, "eV")

    fixed_cuts = [
        fix_one_fermi_edge(cut, edge_fit) for cut in stacked.transpose("flattened", ...)
    ]
    fixed_stack: xr.DataArray = xr.concat(fixed_cuts, dim="flattened")
    # Have to replace only the values to unstack later
    stacked.values = fixed_stack.transpose(*stacked.dims)

    if isinstance(dataset, xr.Dataset):
        return dataset.update({dataset.S.spectrum_name: stacked.unstack("flattened")})
    return stacked.unstack("flattened")



"""An interactive band selection tool used to initialize curve fits."""

import numpy as np
from bokeh import events

import xarray as xr
from arpes.analysis.band_analysis import fit_patterned_bands
from arpes.exceptions import AnalysisError
from arpes.models import band
from arpes.plotting.interactive_utils import CursorTool, SaveableTool
from arpes.utilities import normalize_to_spectrum

__all__ = ("BandTool",)


class BandTool(SaveableTool, CursorTool):
    """Two dimensional fitting band tool."""

    auto_zero_nans = False
    auto_rebin = False

    def __init__(self, **kwargs):
        """Load plot sizes and standard settings from user overrides."""
        super().__init__(kwargs.pop("name", None))

        self.load_settings(**kwargs)

        self.app_main_size = self.settings.get("main_width", 600)
        self.app_marginal_size = self.settings.get("marginal_width", 300)
        self.active_band = None
        self.pointer_mode = "band"

    def tool_handler(self, doc):
        """Sets up widgets for the Bokeh application."""
        from bokeh.layouts import row, column
        from bokeh.models.mappers import LinearColorMapper
        from bokeh.models import widgets
        from bokeh.plotting import figure

        if len(self.arr.shape) != 2:
            raise AnalysisError("Cannot use the band tool on non image-like spectra")

        arr = self.arr
        x_coords, y_coords = arr.coords[arr.dims[0]], arr.coords[arr.dims[1]]

        default_palette = self.default_palette

        self.app_context.update(
            {
                "bands": {},
                "center_float": None,
                "data": arr,
                "data_range": {
                    "x": (np.min(x_coords.values), np.max(x_coords.values)),
                    "y": (np.min(y_coords.values), np.max(y_coords.values)),
                },
                "direction_normal": True,
                "fit_mode": "mdc",
            }
        )

        figures, plots, app_widgets = (
            self.app_context["figures"],
            self.app_context["plots"],
            self.app_context["widgets"],
        )
        self.cursor = [np.mean(self.data_range["x"]), np.mean(self.data_range["y"])]

        self.color_maps["main"] = LinearColorMapper(
            default_palette,
            low=np.min(arr.values),
            high=np.max(arr.values),
            nan_color="black",
        )

        main_tools = ["wheel_zoom", "tap", "reset"]
        main_title = "Band Tool: WARNING Unidentified"

        try:
            main_title = "Band Tool: {}".format(arr.S.label[:60])
        except:
            pass

        figures["main"] = figure(
            tools=main_tools,
            width=self.app_main_size,
            height=self.app_main_size,
            min_border=10,
            min_border_left=50,
            toolbar_location="left",
            x_axis_location="below",
            y_axis_location="right",
            title=main_title,
            x_range=self.data_range["x"],
            y_range=self.data_range["y"],
        )
        figures["main"].xaxis.axis_label = arr.dims[0]
        figures["main"].yaxis.axis_label = arr.dims[1]
        figures["main"].toolbar.logo = None
        figures["main"].background_fill_color = "#fafafa"
        plots["main"] = figures["main"].image(
            [arr.values.T],
            x=self.app_context["data_range"]["x"][0],
            y=self.app_context["data_range"]["y"][0],
            dw=self.app_context["data_range"]["x"][1]
            - self.app_context["data_range"]["x"][0],
            dh=self.app_context["data_range"]["y"][1]
            - self.app_context["data_range"]["y"][0],
            color_mapper=self.app_context["color_maps"]["main"],
        )

        # add lines
        self.add_cursor_lines(figures["main"])
        band_lines = figures["main"].multi_line(
            xs=[], ys=[], line_color="white", line_width=1
        )

        def append_point_to_band():
            cursor = self.cursor
            if self.active_band in self.app_context["bands"]:
                self.app_context["bands"][self.active_band]["points"].append(
                    list(cursor)
                )
                update_band_display()

        def click_main_image(event):
            self.cursor = [event.x, event.y]
            if self.pointer_mode == "band":
                append_point_to_band()

        update_main_colormap = self.update_colormap_for("main")

        POINTER_MODES = [
            ("Cursor", "cursor"),
            ("Band", "band"),
        ]

        FIT_MODES = [
            ("EDC", "edc"),
            ("MDC", "mdc"),
        ]

        DIRECTIONS = [
            ("From Bottom/Left", "forward"),
            ("From Top/Right", "reverse"),
        ]

        BAND_TYPES = [
            ("Lorentzian", "Lorentzian"),
            ("Voigt", "Voigt"),
            ("Gaussian", "Gaussian"),
        ]

        band_classes = {
            "Lorentzian": band.Band,
            "Gaussian": band.BackgroundBand,
            "Voigt": band.VoigtBand,
        }

        self.app_context["band_options"] = []

        def pack_bands():
            packed_bands = {}
            for band_name, band_description in self.app_context["bands"].items():
                if not band_description["points"]:
                    raise AnalysisError("Band {} is empty.".format(band_name))

                stray = None
                try:
                    stray = float(band_description["center_float"])
                except (KeyError, ValueError, TypeError):
                    try:
                        stray = float(self.app_context["center_float"])
                    except Exception:
                        pass

                packed_bands[band_name] = {
                    "name": band_name,
                    "band": band_classes.get(band_description["type"], band.Band),
                    "dims": self.arr.dims,
                    "params": {
                        "amplitude": {"min": 0},
                    },
                    "points": band_description["points"],
                }

                if stray is not None:
                    packed_bands[band_name]["params"]["stray"] = stray

            return packed_bands

        def fit(override_data=None):
            packed_bands = pack_bands()
            dims = list(self.arr.dims)
            if "eV" in dims:
                dims.remove("eV")
            angular_direction = dims[0]
            if isinstance(override_data, xr.Dataset):
                override_data = normalize_to_spectrum(override_data)
            return fit_patterned_bands(
                override_data if override_data is not None else self.arr,
                packed_bands,
                fit_direction=(
                    "eV" if self.app_context["fit_mode"] == "edc" else angular_direction
                ),
                direction_normal=self.app_context["direction_normal"],
            )

        self.app_context["pack_bands"] = pack_bands
        self.app_context["fit"] = fit

        self.pointer_dropdown = widgets.Dropdown(
            label="Pointer Mode", button_type="primary", menu=POINTER_MODES
        )
        self.direction_dropdown = widgets.Dropdown(
            label="Fit Direction", button_type="primary", menu=DIRECTIONS
        )
        self.band_dropdown = widgets.Dropdown(
            label="Active Band",
            button_type="primary",
            menu=self.app_context["band_options"],
        )
        self.fit_mode_dropdown = widgets.Dropdown(
            label="Mode", button_type="primary", menu=FIT_MODES
        )
        self.band_type_dropdown = widgets.Dropdown(
            label="Band Type", button_type="primary", menu=BAND_TYPES
        )

        self.band_name_input = widgets.TextInput(placeholder="Band name...")
        self.center_float_widget = widgets.TextInput(placeholder="Center Constraint")
        self.center_float_copy = widgets.Button(label="Copy to all...")
        self.add_band_button = widgets.Button(label="Add Band")

        self.clear_band_button = widgets.Button(label="Clear Band")
        self.remove_band_button = widgets.Button(label="Remove Band")

        self.main_color_range_slider = widgets.RangeSlider(
            start=0,
            end=100,
            value=(
                0,
                100,
            ),
            title="Color Range",
        )

        def add_band(band_name):
            if band_name not in self.app_context["bands"]:
                self.app_context["band_options"].append(
                    (
                        band_name,
                        band_name,
                    )
                )
                self.band_dropdown.menu = self.app_context["band_options"]
                self.app_context["bands"][band_name] = {
                    "type": "Lorentzian",
                    "points": [],
                    "name": band_name,
                    "center_float": None,
                }

                if self.active_band is None:
                    self.active_band = band_name

                self.save_app()

        def on_copy_center_float():
            for band_name in self.app_context["bands"].keys():
                self.app_context["bands"][band_name]["center_float"] = self.app_context[
                    "center_float"
                ]
                self.save_app()

        def on_change_active_band(event):
            self.app_context["active_band"] = event.item
            self.active_band = event.item

        def on_change_pointer_mode(event):
            self.app_context["pointer_mode"] = event.item
            self.pointer_mode = event.item

        def set_center_float_value(event):
            new_value = event.item
            self.app_context["center_float"] = new_value
            if self.active_band in self.app_context["bands"]:
                self.app_context["bands"][self.active_band]["center_float"] = new_value

            self.save_app()

        def set_fit_direction(event):
            new_direction = event.item
            self.app_context["direction_normal"] = new_direction == "forward"
            self.save_app()

        def set_fit_mode(event):
            new_mode = event.item
            self.app_context["fit_mode"] = new_mode
            self.save_app()

        def set_band_type(event):
            if self.active_band in self.app_context["bands"]:
                self.app_context["bands"][self.active_band]["type"] = event.item

            self.save_app()

        def update_band_display():
            band_names = self.app_context["bands"].keys()
            band_lines.data_source.data = {
                "xs": [
                    [p[0] for p in self.app_context["bands"][b]["points"]]
                    for b in band_names
                ],
                "ys": [
                    [p[1] for p in self.app_context["bands"][b]["points"]]
                    for b in band_names
                ],
            }
            self.save_app()

        self.update_band_display = update_band_display

        def on_clear_band():
            if self.active_band in self.app_context["bands"]:
                self.app_context["bands"][self.active_band]["points"] = []
                update_band_display()

        def on_remove_band():
            if self.active_band in self.app_context["bands"]:
                del self.app_context["bands"][self.active_band]
                new_band_options = [
                    b
                    for b in self.app_context["band_options"]
                    if b[0] != self.active_band
                ]
                self.band_dropdown.menu = new_band_options
                self.app_context["band_options"] = new_band_options
                self.active_band = None
                update_band_display()

        # Attach callbacks
        self.main_color_range_slider.on_change("value", update_main_colormap)

        figures["main"].on_event(events.Tap, click_main_image)
        self.band_dropdown.on_click(on_change_active_band)
        self.pointer_dropdown.on_click(on_change_pointer_mode)
        self.add_band_button.on_click(lambda: add_band(self.band_name_input.value))
        self.clear_band_button.on_click(on_clear_band)
        self.remove_band_button.on_click(on_remove_band)
        self.center_float_copy.on_click(on_copy_center_float)
        # J: broken
        # self.center_float_widget.on_change(set_center_float_value)
        self.direction_dropdown.on_click(set_fit_direction)
        self.fit_mode_dropdown.on_click(set_fit_mode)
        self.band_type_dropdown.on_click(set_band_type)

        layout = row(
            column(figures["main"]),
            column(
                column(
                    self.pointer_dropdown,
                    self.band_dropdown,
                    self.fit_mode_dropdown,
                    self.band_type_dropdown,
                    self.direction_dropdown,
                ),
                row(
                    self.band_name_input,
                    self.add_band_button,
                ),
                row(
                    self.clear_band_button,
                    self.remove_band_button,
                ),
                row(self.center_float_widget, self.center_float_copy),
                column(
                    self._cursor_info,
                    self.main_color_range_slider,
                ),
            ),
        )

        doc.add_root(layout)
        doc.title = "Band Tool"
        self.load_app()
        self.save_app()

    def serialize(self):
        """Saves application state so it can be recovered later.

        See `.deserialize` for information on what is retained.
        """
        return {
            "active_band": self.active_band,
            "band_options": self.band_options,
            "bands": self.bands,
            "center_float": self.app_context["center_float"],
            "cursor": self.cursor,
            "direction_normal": self.direction_normal,
            "fit_mode": self.fit_mode,
        }

    def deserialize(self, json_data):
        """Loads a variety of application state so it can be recovered later.

        Mostly we load:
        * Band definitions
        * Modes for fitting
        * The cursor location
        """
        self.cursor = json_data.get("cursor", [0, 0])

        # self.active_band = json_data.get('active_band', None)

        self.app_context["center_float"] = json_data.get("center_float", None)
        self.app_context["bands"] = json_data.get("bands", {}) or {}
        self.app_context["direction_normal"] = json_data.get("direction_normal", True)
        self.app_context["fit_mode"] = json_data.get("fit_mode", "mdc")
        self.app_context["band_options"] = json_data.get("band_options", [])

        self.center_float_widget.value = str(self.app_context["center_float"] or 0.1)
        self.band_dropdown.menu = self.app_context["band_options"]

        self.update_band_display()



from arpes.utilities.qt.utils import PlotOrientation
from PyQt5 import QtGui, QtWidgets
import numpy as np
import xarray as xr
import weakref
from typing import override
from collections import defaultdict


from arpes.utilities.qt.ui import (
    KeyBinding,
    Key,
    KeyboardModifier,
    horizontal,
    vertical,
    tabs,
)
from arpes.utilities.qt import (
    qt_info,
    BasicHelpDialog,
    SimpleWindow,
    SimpleApp,
)

from .AxisInfoWidget import AxisInfoWidget
from .BinningInfoWidget import BinningInfoWidget

__all__ = (
    "ImageTool",
    "image_tool",
)

qt_info.setup_pyqtgraph()


class ImageToolWindow(SimpleWindow):
    """
    The application window for `QtTool`.

    QtToolWindow was the first Qt-Based Tool that I built for PyARPES. Much of its
    structure was ported to SimpleWindow and borrowed ideas from when I wrote DAQuiri.
    As a result, the structure is essentially now to define just the handlers and any
    lifecycle hooks (close, etc.)
    """

    HELP_DIALOG_CLS = BasicHelpDialog
    SCROLL_INCREMENT = 5
    INCREMENT_MULTIPLIER = 5

    def compile_key_bindings(self):
        return super().compile_key_bindings() + [  # already includes Help and Close
            KeyBinding(
                "Scroll Cursor",
                [
                    Key.Key_Left,
                    Key.Key_Right,
                    Key.Key_Up,
                    Key.Key_Down,
                ],
                self.scroll,
            ),
            KeyBinding(
                "Reset Intensity",
                [Key.Key_I],
                self.reset_intensity,
            ),
            KeyBinding(
                "Scroll Z-Cursor",
                [
                    Key.Key_N,
                    Key.Key_M,
                ],
                self.scroll_z,
            ),
            KeyBinding(
                "Center Cursor",
                [Key.Key_C],
                self.center_cursor,
            ),
            KeyBinding(
                "Transpose - Roll Axis",
                [Key.Key_Y],
                self.transpose_roll,
            ),
            KeyBinding(
                "Transpose - Swap Front Axes",
                [Key.Key_T],
                self.transpose_swap,
            ),
        ]

    def center_cursor(self, event):
        self.app().center_cursor()

    def transpose_roll(self, event):
        self.app().transpose_to_front(-1)

    def transpose_swap(self, event):
        self.app().transpose_to_front(1)

    def reset_intensity(self, event: QtGui.QKeyEvent):
        self.app().reset_intensity()

    def get_increment(self, event: QtGui.QKeyEvent):
        if event.modifiers() & KeyboardModifier.ShiftModifier:
            return self.SCROLL_INCREMENT * self.INCREMENT_MULTIPLIER
        elif event.modifiers() & KeyboardModifier.ControlModifier:
            return self.SCROLL_INCREMENT // self.INCREMENT_MULTIPLIER
        return self.SCROLL_INCREMENT

    def scroll_z(self, event: QtGui.QKeyEvent):
        key_map = {
            Key.Key_N: (2, -1),
            Key.Key_M: (2, 1),
        }

        dim_i, delta = key_map.get(event.key())
        delta *= self.get_increment(event)
        if delta is not None and self.app() is not None:
            self.app().scroll(dim_i, delta)

    def scroll(self, event: QtGui.QKeyEvent):
        key_map = {
            Key.Key_Left: (0, -1),
            Key.Key_Right: (0, 1),
            Key.Key_Down: (1, -1),
            Key.Key_Up: (1, 1),
        }

        dim_i, delta = key_map.get(event.key())
        delta *= self.get_increment(event)
        if delta is not None and self.app() is not None:
            self.app().scroll(dim_i, delta)


class ImageTool(SimpleApp):
    """QtTool is an implementation of Image/Bokeh Tool based on PyQtGraph and PyQt5.

    For now we retain a number of the metaphors from BokehTool, including a "context"
    that stores the state, and can be used to programmatically interface with the tool.
    """

    TITLE = "Image Tool"
    WINDOW_CLS = ImageToolWindow
    WINDOW_SIZE = (12, 10)

    def __init__(self):
        """Initialize attributes to safe empty values."""
        super().__init__()
        self.data = None

        self.content_layout = None
        self.main_layout = None

        self.axis_info_widgets = []
        self.binning_info_widgets = []
        self.kspace_info_widgets = []

        self._binning = None

    @override
    def update_cursor_position(self, new_cursor, **kwargs):
        super().update_cursor_position(new_cursor, **kwargs)

        for widget in self.axis_info_widgets + self.binning_info_widgets:
            widget.recompute()

    def center_cursor(self):
        """Scrolls so that the cursors are in the center of the data volume."""
        centers = {dim: self.spectrum.sizes[dim] / 2 for dim in self.spectrum.dims}
        new_cursor = {
            dim: (center - self._binning[dim] / 2, center + self._binning[dim] / 2)
            for dim, center in centers.items()
        }

        for dim, cursors in self.registered_cursors.items():
            for cursor in cursors:
                cursor.setRegion(new_cursor[dim])

    def scroll(self, dim_i: int, delta: int):
        """Scroll the cursor with dim specified by dim_i by delta."""
        dim = self.spectrum.dims[dim_i]
        cursor = self.context["cursor"].copy()
        new_cursor_position = tuple(
            old_position + delta for old_position in cursor[dim]
        )

        def clamp(positions: tuple[float, float], dim: str):
            start, end = positions
            if start < 0:
                start = 0
                end = self._binning[dim]
            elif end > self.spectrum.sizes[dim]:
                end = self.spectrum.sizes[dim]
                start = end - self._binning[dim]
            return start, end

        cursor[dim] = clamp(new_cursor_position, dim)
        for dim_cursor in self.registered_cursors[dim]:
            dim_cursor.setRegion(cursor[dim])

    @property
    def binning(self):
        """The binning on each axis in number of indices."""
        if self._binning is None:
            return {dim: 1 for dim in self.spectrum.dims}

        return self._binning

    @binning.setter
    def binning(self, values: dict[str, int]):
        """Set the desired axis binning."""
        different_binnings = {
            dim: n_bins
            for dim, n_bins in values.items()
            if n_bins != self._binning[dim]
        }
        self._binning = values

        for dim in different_binnings:
            cursors = self.registered_cursors[dim]
            for cursor in cursors:
                cursor.set_width(self._binning[dim])

    def change_cursor_dims(self, dim_mapping: dict[str, str]) -> None:
        """Change the dimensions of the cursors according to the provided mapping."""
        old_cursors = self.registered_cursors
        self.registered_cursors = defaultdict(list)
        for old_dim, cursor_list in old_cursors.items():
            new_dim = dim_mapping[old_dim]
            new_bounds = [0, self.spectrum.sizes[new_dim]]
            new_width = self.binning[new_dim]
            for cursor in cursor_list:
                cursor.setBounds(new_bounds)
                cursor.set_width(new_width)
                self.reconnect_cursor(new_dim, cursor)

    def transpose(self, transpose_order: list[str]):
        """Transpose dimensions into the specified order and redraw."""
        original_order = list(self.spectrum.dims)
        dim_mapping = {
            original_dim: new_dim
            for original_dim, new_dim in zip(original_order, transpose_order)
        }
        self.spectrum = self.spectrum.transpose(*transpose_order, ...)

        for reactive_view in self.reactive_views:
            old_dims = reactive_view.dims
            reactive_view.dims = tuple(dim_mapping[dim] for dim in old_dims)

        for widget in self.axis_info_widgets + self.binning_info_widgets:
            widget.recompute()

        self.change_cursor_dims(dim_mapping)
        cursor_positions = self.context["cursor"]
        for dim, cursors in self.registered_cursors.items():
            for cursor in cursors:
                cursor.setRegion(cursor_positions[dim])

        self.update_reactive_views(force=True, keep_levels=False)

    def transpose_to_front(self, dim: str | int):
        """Transpose dim to the front so that it is in the main marginal."""
        dim = dim if isinstance(dim, str) else self.spectrum.dims[dim]

        order = list(self.spectrum.dims)
        order.remove(dim)
        order = [dim] + order
        self.transpose(order)

    def configure_image_widgets(self):
        """
        Configure array marginals for the input data.

        Depending on the array dimensionality, we need a different number and variety
        of marginals. This is as easy as specifying which marginals we select over and
        handling the rest dynamically.

        An additional complexity is that we also handle the cursor registration here.
        """
        dims = self.spectrum.dims
        if len(dims) == 2:
            self.generate_marginal_for(
                (), 1, 0, "xy", cursors=True, layout=self.content_layout
            )
            self.generate_marginal_for(
                (dims[1],),
                0,
                0,
                "x",
                orientation=PlotOrientation.Horizontal,
                layout=self.content_layout,
            )
            self.generate_marginal_for(
                (dims[0],),
                1,
                1,
                "y",
                orientation=PlotOrientation.Vertical,
                layout=self.content_layout,
            )

            self.views["xy"].view.setYLink(self.views["y"])
            self.views["xy"].view.setXLink(self.views["x"])

        if len(dims) == 3:
            self.generate_marginal_for(
                (dims[1], dims[2]),
                0,
                0,
                "x",
                orientation=PlotOrientation.Horizontal,
                layout=self.content_layout,
            )
            self.generate_marginal_for(
                (dims[1],), 1, 0, "xz", cursors=True, layout=self.content_layout
            )
            self.generate_marginal_for(
                (dims[2],), 2, 0, "xy", cursors=True, layout=self.content_layout
            )
            self.generate_marginal_for(
                (dims[0], dims[1]),
                1,
                1,
                "z",
                orientation=PlotOrientation.Horizontal,
                layout=self.content_layout,
            )
            self.generate_marginal_for(
                (dims[0], dims[2]),
                2,
                2,
                "y",
                orientation=PlotOrientation.Vertical,
                layout=self.content_layout,
            )
            self.generate_marginal_for(
                (dims[0],), 2, 1, "yz", cursors=True, layout=self.content_layout
            )

            self.views["xy"].view.setYLink(self.views["y"])
            self.views["xy"].view.setXLink(self.views["x"])
            self.views["xz"].view.setXLink(self.views["z"])
            self.views["xz"].view.setXLink(self.views["xy"].view)

        if len(dims) == 4:
            self.generate_marginal_for(
                (dims[1], dims[3]), 0, 0, "xz", layout=self.content_layout
            )
            self.generate_marginal_for(
                (dims[2], dims[3]), 1, 0, "xy", cursors=True, layout=self.content_layout
            )
            self.generate_marginal_for(
                (dims[0], dims[2]), 1, 1, "yz", layout=self.content_layout
            )
            self.generate_marginal_for(
                (dims[0], dims[1]), 0, 1, "zw", cursors=True, layout=self.content_layout
            )

    def construct_axes_tab(self):
        """Controls for axis order and transposition."""
        inner_items = [
            AxisInfoWidget(axis_name=axis_name, root=weakref.ref(self))
            for axis_name in self.spectrum.dims
        ]
        return horizontal(*inner_items), inner_items

    def construct_binning_tab(self):
        """This tab controls the degree of binning around the cursor."""
        inner_items = [
            BinningInfoWidget(axis_name=dim, root=weakref.ref(self))
            for dim in self.spectrum.dims
        ]

        return horizontal(*inner_items), inner_items

    # TODO
    def construct_color_tab(self): ...

    def construct_help_tab(self):
        """The help tab."""
        texts = [
            "Controls:",
            "   Move main cursor - ↓ ↑ ← →, Shift for 5x, Ctrl for 1/5x",
            "   Move Z-cursor - n/m, Shift for 5x, Ctrl for 1/5x",
            "   Reset intensity - i    Center cursor - c",
            "   Transpose Main - t    Swap Main - y",
        ]
        return vertical(*[QtWidgets.QLabel(t) for t in texts])

    def add_contextual_widgets(self):
        """Adds the widgets for the contextual controls at the bottom."""
        axes_tab, self.axis_info_widgets = self.construct_axes_tab()
        binning_tab, self.binning_info_widgets = self.construct_binning_tab()

        self.tabs = tabs(
            ["Binning", binning_tab],
            ["Axes", axes_tab],
            ["Help", self.construct_help_tab()],
        )
        self.tabs.setFixedHeight(qt_info.inches_to_px(2))

        self.main_layout.addLayout(self.content_layout, 0, 0)
        self.main_layout.addWidget(self.tabs, 1, 0)

    def layout(self):
        """Initialize the layout components."""
        self.main_layout = QtWidgets.QGridLayout()
        self.content_layout = QtWidgets.QGridLayout()
        return self.main_layout

    def before_show(self):
        """Lifecycle hook for configuration before app show."""
        self.configure_image_widgets()
        self.add_contextual_widgets()
        from matplotlib import colormaps

        if self.spectrum.min() >= 0.0:
            self.set_colormap(colormaps["magma"])
        else:
            self.set_colormap(colormaps["RdBu_r"])

    def after_show(self):
        """Initialize application state after app show.

        To do this, we need to set the initial cursor location, and call update
        which forces a rerender.
        """
        # basic state initialization
        self.context.update(
            {
                "cursor": {dim: (0, 1) for dim in self.spectrum.dims},
            }
        )

        # Display the data
        self.center_cursor()
        self.reset_intensity()

    def reset_intensity(self):
        """Autoscales intensity in each marginal plot."""
        self.update_reactive_views(force=True, keep_levels=False)

    def set_data(self, data: xr.Dataset):
        """Sets the current data to a new value and resets binning."""

        if np.any(np.isnan(data)):
            data = data.fillna(0)

        self.data = data
        self._binning = {dim: 1 for dim in self.spectrum.dims}
